#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Nov  1 20:08:25 2022

@author: alex
"""

import numpy as np

from pyro.dynamic import massspringdamper



m = 1
b = 0

sys = massspringdamper.FloatingSingleMass( m , b )

sys.plot_phase_plane()

sys.x0 = np.array([0,5])
sys.compute_trajectory( tf = 1 )
sys.plot_phase_plane_trajectory()
sys.animate_simulation()


m = 1
b = 0.1

sys = massspringdamper.FloatingSingleMass( m , b )

sys.plot_phase_plane()



m = 1
b = 2

sys = massspringdamper.FloatingSingleMass( m , b )

sys.plot_phase_plane()

sys.x0 = np.array([0,5])
sys.compute_trajectory( tf = 1 )
sys.plot_phase_plane_trajectory()
sys.animate_simulation()


# Pendulum

from pyro.dynamic import pendulum

sys = pendulum.SinglePendulum()

sys.plot_phase_plane()


sys.x0 = np.array([-6.28,5])
sys.compute_trajectory( tf = 3 )
sys.plot_phase_plane_trajectory()
sys.animate_simulation()

sys.x0 = np.array([0,1])
sys.compute_trajectory( tf = 5 )
sys.plot_phase_plane_trajectory()
sys.animate_simulation()


sys.d1 = 1

sys.plot_phase_plane()

sys.x0 = np.array([0,5])
sys.compute_trajectory( tf = 10 )
sys.plot_phase_plane_trajectory()
sys.animate_simulation()



# Custom

from pyro.dynamic import system


class CustomSys( system.ContinuousDynamicSystem ):
    
    ############################
    def __init__(self):
        
        # initialize standard 2 state (n=2) dynamic system
        system.ContinuousDynamicSystem.__init__( self, n = 2 )
        
        # Name and labels
        self.name = 'My custom system'
        self.state_label = [ 'State 1' , 'State 2' ]
        self.state_units = [ '[]', '[]']

    
    #############################
    def f( self , x , u , t ):
        """ 
        Continuous time foward dynamics evaluation dx = f(x,u,t)
        
        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        
        OUPUTS
        dx : state derivative vector  n x 1
        
        """
        
        dx = np.zeros(self.n) # State derivative vector
        
        # Your Dynamic Equation bellow:
        dx[0] = -x[0]
        dx[1] = -x[1]
        
        return dx



sys = CustomSys()

sys.plot_phase_plane()


# Closed-loop

m = 1
b = 0

sys = massspringdamper.FloatingSingleMass( m , b )

sys.p = 2
sys.C = np.diag([1,1])
sys.cost_function = None

from pyro.control import linear

ctl = linear.ProportionalController( 1 , 2)


ctl.K[0,0] = 1.0
ctl.K[0,1] = 0.5

ctl.rbar = np.array([2,0])

cl_sys = ctl + sys

cl_sys.plot_phase_plane()

cl_sys.plot_phase_plane_closed_loop()



cl_sys.x0 = np.array([0,5])
cl_sys.compute_trajectory( tf = 10 )
cl_sys.plot_phase_plane_trajectory()
cl_sys.animate_simulation()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Oct 14 20:48:32 2022

@author: alex
"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation

import time

from scipy.interpolate import RectBivariateSpline as interpol2D
from scipy.interpolate import RegularGridInterpolator as rgi



from pyro.control  import controller


###############################################################################
### DP controllers
###############################################################################

class LookUpTableController( controller.StaticController ):

    ############################
    def __init__(self, grid_sys , pi ):
        """
        Pyro controller based on a discretized lookpup table of control inputs

        Parameters
        ----------
        grid_sys : pyro GridDynamicSystem class
            A discretized dynamic system
        pi : numpy array, dim =  self.grid_sys.nodes_n , dtype = int
            A list of action index for each node id
        """
        
        if grid_sys.nodes_n != pi.size:
            raise ValueError("Grid size does not match optimal action table size")
        
        k = 1                   # Ref signal dim
        m = grid_sys.sys.m      # control input signal dim
        p = grid_sys.sys.n      # output signal dim (state feedback)
        
        super().__init__(k, m, p)
        
        # Grid sys
        self.grid_sys = grid_sys
        
        # Table of actions
        self.pi = pi
        
        # Label
        self.name = 'Tabular Controller'
        
        # Interpolation Options
        self.interpol_method = []
        
        for k in range(self.m):
            
            # options can be changed for each control input axis
            self.interpol_method.append('linear') # "linear”, “nearest”, “slinear”, “cubic”, and “quintic”
            
        self.compute_interpol_functions()
        
    
    #############################
    def compute_interpol_functions( self  ):
        """  """
        
        self.u_interpol = [] 
        
        for k in range(self.m):
            
            u_k      = self.grid_sys.get_input_from_policy( self.pi, k)
            self.u_interpol.append( self.grid_sys.compute_interpolation_function( u_k , 
                                                                                 self.interpol_method[k],
                                                                                 bounds_error = False   , 
                                                                                 fill_value = 0  ) )
        
    
    #############################
    def lookup_table_selection( self , x ):
        """  select the optimal u given actual cost map """
        
        u = np.zeros( self.m )
        
        for k in range(self.m):
            
            u[k] = self.u_interpol[k]( x )
            
        return u
    

    #############################
    def c( self , y , r , t = 0 ):
        """  State feedback (y=x) - no reference - time independent """
        
        x = y
        
        u = self.lookup_table_selection( x )
        
        return u
    
    

###############################################################################
### DP Algo
###############################################################################

class DynamicProgramming:
    """ Dynamic programming on a grid sys """
    
    ############################
    def __init__(self, grid_sys , cost_function , final_time = 0 ):
        
        # Dynamic system
        self.grid_sys  = grid_sys         # Discretized Dynamic system class
        self.sys       = grid_sys.sys     # Base Dynamic system class
        
        # Cost function
        self.cf  = cost_function
        self.tf  = final_time
        
        # Options
        self.alpha                = 1.0     # exponential forgetting factor
        self.interpol_method      ='linear' # "linear”, “nearest”, “slinear”, “cubic”, and “quintic”
        self.save_time_history    = True
        
        # Memory Variables
        self.t = self.tf   # time of the computed step (useful for time-varying system)
        self.k = 0         # Number of computed steps
        
        
        # Start time (needed to plot elapsed computation time)
        self.start_time = time.time()
        
        # Final cost
        self.evaluate_terminal_cost()
        
        # 
        if self.save_time_history:

            self.t_list  = []
            self.J_list  = []
            self.pi_list = []
            
            # Value at t = t_f
            self.J_list.append(  self.J  )
            self.t_list.append(  self.tf )
            self.pi_list.append( self.pi )
        
        
    ##############################
    def evaluate_terminal_cost(self):
        """ initialize cost-to-go and policy """

        self.J  = np.zeros( self.grid_sys.nodes_n , dtype = float )
        self.pi = np.zeros( self.grid_sys.nodes_n , dtype = int   )

        # Initial cost-to-go evaluation       
        for s in range( self.grid_sys.nodes_n ):  
            
                xf = self.grid_sys.state_from_node_id[ s , : ]
                
                # Final Cost of all states
                self.J[ s ] = self.cf.h( xf , self.tf )
                
    
    ###############################
    def initialize_backward_step(self):
        """ One step of value iteration """
        
        # Update values
        self.k      = self.k + 1                  # index backward in time
        self.t      = self.t - self.grid_sys.dt   # time
        self.J_next = self.J
        
        # New Cost-to-go and policy array to be computed
        self.J  = np.zeros( self.grid_sys.nodes_n , dtype = float )
        self.pi = np.zeros( self.grid_sys.nodes_n , dtype = int   )
        
        # Create interpol function
        self.J_interpol = self.grid_sys.compute_interpolation_function( self.J_next               , 
                                                                        self.interpol_method      , 
                                                                        bounds_error = False      , 
                                                                        fill_value = 0  )
                        
                
    ###############################
    def compute_backward_step(self):
        """ One step of value iteration """
        
        # For all state nodes        
        for s in range( self.grid_sys.nodes_n ):  
            
                x = self.grid_sys.state_from_node_id[ s , : ]
                
                Q = np.zeros( self.grid_sys.actions_n  ) 
                
                # For all control actions
                for a in range( self.grid_sys.actions_n ):
                    
                    u = self.grid_sys.input_from_action_id[ a , : ]                  
                        
                    # If action is in allowable set
                    if self.sys.isavalidinput( x , u ):
                        
                        # Forward dynamics 
                        x_next = self.sys.f( x , u , self.t ) * self.grid_sys.dt + x
                        
                        # if the next state is not out-of-bound
                        if self.sys.isavalidstate(x_next):

                            # Estimated (interpolation) cost to go of arrival x_next state
                            J_next = self.J_interpol( x_next )
                            
                            # Cost-to-go of a given action
                            Q[ a ] = self.cf.g( x , u , self.t ) * self.grid_sys.dt + self.alpha * J_next
                            
                        else:
                            
                            # Out of bound terminal cost
                            Q[ a ] = self.cf.INF # TODO add option to customize this
                        
                    else:
                        
                        # Invalide control input at this state
                        Q[ a ] = self.cf.INF
                        
                self.J[ s ]  = Q.min()
                self.pi[ s ] = Q.argmin()
                    
    
    ###############################
    def finalize_backward_step(self):
        """ One step of value iteration """
        
        # Computation time
        elapsed_time = time.time() - self.start_time
        
        # Convergence check        
        delta     = self.J - self.J_next
        j_max     = self.J.max()
        delta_max = delta.max()
        delta_min = delta.min()
        
        print('%d t:%.2f Elasped:%.2f max: %.2f dmax:%.2f dmin:%.2f' % (self.k,self.t,elapsed_time,j_max,delta_max,delta_min) )
        
        # List in memory
        if self.save_time_history:
            self.J_list.append(  self.J  )
            self.t_list.append(  self.t  )
            self.pi_list.append( self.pi )
            
        # return largest J change for usage as stoping criteria
        return abs(np.array([delta_max,delta_min])).max() 

    
    ################################
    def compute_steps(self, n = 50 , animate_cost2go = False , animate_policy = False , k = 0 ):
        """ compute number of step """
        
        print('\nComputing %d backward DP iterations:'%n)
        print('-----------------------------------------')
        
        if animate_cost2go: self.plot_cost2go()
        if animate_policy:  self.plot_policy( k )
               
        for i in range(n):
            self.initialize_backward_step()
            self.compute_backward_step()
            self.finalize_backward_step()
            if animate_cost2go: self.update_cost2go_plot()
            if animate_policy:  self.update_policy_plot( k )
            
    
    ################################
    def solve_bellman_equation(self, tol = 0.1 , animate_cost2go = False , animate_policy = False , k = 0 ):
        """ 
        Value iteration algorithm
        --------------------------
        
        Do Dp backward iterations until changes to J are under the tolerance 
        
        Note: self.alpha should be smaller then 1 to garantee convergence
        
        
        """
        
        print('\nComputing backward DP iterations until dJ<%2.2f:'%tol)
        print('---------------------------------------------------------')
        
        #self.plot_cost2go()
        #self.plot_policy( k )
        if animate_cost2go: self.plot_cost2go()
        if animate_policy: self.plot_policy( k )
        
        delta = self.cf.INF
        
        while (delta>tol):
            self.initialize_backward_step()
            self.compute_backward_step()
            delta = self.finalize_backward_step()
            #self.update_cost2go_plot()
            #self.update_policy_plot( k )
            if animate_cost2go: self.update_cost2go_plot()
            if animate_policy: self.update_policy_plot( k )
            
        print('\nBellman equation solved!' )
        
        
    ################################
    ### Data tools
    ################################
    
    ################################
    def clean_infeasible_set(self , tol = 1):
        """
        Set default policy and cost2go to cf.INF for state for  which it is unavoidable
        that they will reach unallowable regions

        """
        
        default_action = self.grid_sys.get_nearest_action_id_from_input( self.sys.ubar )
        
        infeasible_node_IDs = self.J > ( self.cf.INF - tol )
        
        self.J[  infeasible_node_IDs ] = self.cf.INF
        self.pi[ infeasible_node_IDs ] = default_action
        
        
    ################################
    ### Print quick shorcuts
    ################################
            
            
    ################################
    def plot_cost2go(self , jmax = None , i = 0 , j = 1 , show = True ):
        
        if jmax == None: jmax = self.cf.INF
               
        fig, ax, pcm = self.grid_sys.plot_grid_value( self.J , 'Cost-to-go' , i , j , jmax , 0 )
        
        text = ax.text(0.05, 0.05, '', transform=ax.transAxes, fontsize = 8 )
        
        self.cost2go_fig = [fig, ax, pcm, text, i , j ]
        
        plt.ion()
        if show: plt.pause( 0.001 )
        
        
    ################################
    def update_cost2go_plot(self, show = True ):
        
        J_grid = self.grid_sys.get_grid_from_array( self.J )
        
        J_2d = self.grid_sys.get_2D_slice_of_grid( J_grid , self.cost2go_fig[4] , self.cost2go_fig[5] )
               
        self.cost2go_fig[2].set_array( np.ravel( J_2d.T ) )
        self.cost2go_fig[3].set_text('Optimal cost2go at time = %4.2f' % ( self.t ))
        
        if show: plt.pause( 0.001 )
        
    
    ################################
    def plot_policy(self , k = 0 , i = 0 , j = 1 , show = True ):
               
        fig, ax, pcm = self.grid_sys.plot_control_input_from_policy( self.pi , k , i , j)
        
        text = ax.text(0.05, 0.05, '', transform=ax.transAxes, fontsize = 8 )
        
        self.policy_fig = [fig, ax, pcm, text, k , i , j ]
        
        plt.ion()
        if show: plt.pause( 0.001 )
        
        
    ################################
    def update_policy_plot(self, show = True  ):
        
        uk    = self.grid_sys.get_input_from_policy( self.pi, self.policy_fig[4] )
        uk_nd = self.grid_sys.get_grid_from_array( uk ) 
        uk_2d = self.grid_sys.get_2D_slice_of_grid( uk_nd , self.policy_fig[5] , self.policy_fig[6] )
               
        self.policy_fig[2].set_array( np.ravel( uk_2d.T ) )
        self.policy_fig[3].set_text('Optimal policy at time = %4.2f' % ( self.t ))
        
        if show: plt.pause( 0.001 )
        
        
    ################################
    def animate_cost2go(self, i = 0 , j = 1 , jmax = None , show = True , save = False ,  file_name = 'cost2go_animation'):
        
        self.J  = self.J_list[0]
        #self.pi = self.pi_list[0]
        self.t  = self.t_list[0]
        self.clean_infeasible_set()
        
        self.plot_cost2go( jmax = jmax , i = i , j = j , show = False  )

        self.ani = animation.FuncAnimation( self.cost2go_fig[0], self.__animate_cost2go, 
                                                len(self.J_list), interval = 20 )
        
        if save: self.ani.save( file_name + '.gif', writer='imagemagick', fps=30)
        
        if show: self.cost2go_fig[0].show()
        
        return self.ani
        
    
    ################################
    def __animate_cost2go(self , i ):
        
        self.J  = self.J_list[i]
        #self.pi = self.pi_list[i]
        self.t  = self.t_list[i]
        self.clean_infeasible_set()
        self.update_cost2go_plot( show = False )
        
        
    ################################
    def animate_policy(self , k = 0 , i = 0 , j = 1 , show = True , save = False , file_name = 'policy_animation'):
        
        self.J  = self.J_list[1]
        self.pi = self.pi_list[1]
        self.t  = self.t_list[1]
        self.clean_infeasible_set()
        self.plot_policy( k = k , i = i , j = j , show = False )

        self.ani = animation.FuncAnimation( self.policy_fig[0], self.__animate_policy, 
                                                len(self.pi_list)-1, interval = 20 )
        
        if save: self.ani.save( file_name + '.gif', writer='imagemagick', fps=30)
        
        if show: self.policy_fig[0].show()
        
        return self.ani
        
    
    ################################
    def __animate_policy(self , i ):
        
        self.J  = self.J_list[i+1]
        self.pi = self.pi_list[i+1]
        self.t  = self.t_list[i+1]
        self.clean_infeasible_set()
        self.update_policy_plot( show = False )
        
    
    ################################
    def plot_cost2go_3D(self , jmax = None , i = 0 , j = 1 , show = True ):
        
        if jmax == None: jmax = self.cf.INF
               
        fig, ax, surf = self.grid_sys.plot_grid_value_3D( self.J , None , 'Cost-to-go' , i , j , jmax , 0)
        
        text = ax.text(0.05, 0.05, 0.05, '', transform=ax.transAxes, fontsize = 8 )
        
        self.cost2go_3D_fig = [fig, ax, surf, text, i , j ]
        
    
    ################################
    ### Quick utility shorcuts
    ################################
    
    ################################
    def get_lookup_table_controller(self):
        """ Create a pyro controller object based on the latest policy """
        
        ctl = LookUpTableController( self.grid_sys, self.pi )
        
        return ctl
        
        
    ################################
    def save_latest(self, name = 'test_data'):
        """ save cost2go and policy of the latest iteration (further back in time) """
        
        np.save(name + '_J_inf', self.J_next)
        np.save(name + '_pi_inf', self.pi.astype(int) )
        
    
    ################################
    def load_J_next(self, name = 'test_data'):
        """ Load J_next from file """
        
        try:

            self.J_next = np.load( name + '_J_inf'   + '.npy' )
            #self.pi     = np.load( name + '_pi_inf'  + '.npy' ).astype(int)
            
        except:
            
            print('Failed to load J_next ' )
            

                    
###############################################################################
    
class DynamicProgrammingWithLookUpTable( DynamicProgramming ):
    """ Dynamic programming on a grid sys """
    
    ############################
    def __init__(self, grid_sys , cost_function , final_time = 0 ):
        
        DynamicProgramming.__init__(self, grid_sys, cost_function, final_time)
        
        self.compute_cost_lookuptable()
    
    
    ###############################
    def compute_cost_lookuptable(self):
        """ One step of value iteration """
        
        start_time = time.time()
        print('Computing g(x,u,t) look-up table..  ', end = '')
        
        self.G = np.zeros( ( self.grid_sys.nodes_n , self.grid_sys.actions_n ) , dtype = float )

        # For all state nodes        
        for s in range( self.grid_sys.nodes_n ):  
            
                x = self.grid_sys.state_from_node_id[ s , : ]
                
                # For all control actions
                for a in range( self.grid_sys.actions_n ):
                    
                    # If action is in allowable set
                    if self.grid_sys.action_isok[s,a]:
                        
                        # if the next state is not out-of-bound
                        if self.grid_sys.x_next_isok[s,a]:
                            
                            u = self.grid_sys.input_from_action_id[ a , : ]  
                            
                            self.G[ s , a ] = self.cf.g( x , u , self.t ) * self.grid_sys.dt
                        
                        else:
                            # Out of bound cost (J_interpol return 0 in this case)
                            self.G[ s , a ] = self.cf.INF
                    
                    else:
                        # Not allowable input at this state
                        self.G[ s , a ] = self.cf.INF
        
        # Print update
        computation_time = time.time() - start_time
        print('completed in %4.2f sec'%computation_time)
    
                
    ###############################
    def compute_backward_step(self):
        """ One step of value iteration """
        
        self.Q       = np.zeros( ( self.grid_sys.nodes_n , self.grid_sys.actions_n ) , dtype = float )
        self.Jx_next = np.zeros( ( self.grid_sys.nodes_n , self.grid_sys.actions_n ) , dtype = float )
        
        # Computing the J_next of all x_next in the look-up table
        self.Jx_next = self.J_interpol( self.grid_sys.x_next_table )
        
        # Matrix version of computing all Q values
        self.Q       = self.G + self.alpha * self.Jx_next
                        
        self.J  = self.Q.min( axis = 1 )
        self.pi = self.Q.argmin( axis = 1 )
                

                    
                    

###############################################################################

class DynamicProgramming2DRectBivariateSpline( DynamicProgrammingWithLookUpTable ):
    """ Dynamic programming on a grid sys """
    
    ###############################
    def initialize_backward_step(self):
        """ One step of value iteration """
        
        # Update values
        self.k      = self.k + 1                  # index backward in time
        self.t      = self.t - self.grid_sys.dt   # time
        self.J_next = self.J
        
        # New Cost-to-go and policy array to be computed
        self.J  = np.zeros( self.grid_sys.nodes_n , dtype = float )
        self.pi = np.zeros( self.grid_sys.nodes_n , dtype = int   )
        
        # Create interpol function
        self.J_interpol = self.grid_sys.compute_bivariatespline_2D_interpolation_function( self.J_next , kx=3, ky=3)
    
                
    ###############################
    def compute_backward_step(self):
        """ One step of value iteration """
        
        self.Q       = np.zeros( ( self.grid_sys.nodes_n , self.grid_sys.actions_n ) , dtype = float )
        self.Jx_next = np.zeros( ( self.grid_sys.nodes_n , self.grid_sys.actions_n ) , dtype = float )
        
        # Computing the J_next of all x_next in the look-up table
        X            = self.grid_sys.x_next_table
        Jx_next_flat = self.J_interpol( X[:,:,0].flatten() , X[:,:,1].flatten() , grid = False )
        Jx_next      = Jx_next_flat.reshape( (self.grid_sys.nodes_n , self.grid_sys.actions_n ) )
        
        # Matrix version of computing all Q values
        self.Q       = self.G + self.alpha * Jx_next
                        
        self.J  = self.Q.min( axis = 1 )
        self.pi = self.Q.argmin( axis = 1 )
        
        


###############################################################################
### Policy Evaluation
###############################################################################

class PolicyEvaluator( DynamicProgramming ):
    """ Evaluate the cost2o of a given control law """
    
    ############################
    def __init__(self, ctl , grid_sys , cost_function , final_time = 0 ):
        
        
        DynamicProgramming.__init__(self, grid_sys, cost_function, final_time )
        
        self.ctl = ctl
        
        # Evaluate policy (control law on the grid)
                        
                
    ###############################
    def compute_backward_step(self):
        """ One step of value iteration """
        
        # For all state nodes        
        for s in range( self.grid_sys.nodes_n ):  
            
                x = self.grid_sys.state_from_node_id[ s , : ]
                
                ######################################
                # Action taken by the controller
                r = self.ctl.rbar
                u = self.ctl.c( x , r , self.t )   
                ######################################
                    
                # If action is in allowable set
                if self.sys.isavalidinput( x , u ):
                    
                    # Forward dynamics 
                    x_next = self.sys.f( x , u , self.t ) * self.grid_sys.dt + x
                    
                    # if the next state is not out-of-bound
                    if self.sys.isavalidstate(x_next):

                        # Estimated (interpolation) cost to go of arrival x_next state
                        J_next = self.J_interpol( x_next )
                        
                        # Cost-to-go of a given action
                        Q = self.cf.g( x , u , self.t ) * self.grid_sys.dt + self.alpha * J_next
                        
                    else:
                        
                        # Out of bound terminal cost
                        Q = self.cf.INF # TODO add option to customize this
                    
                else:
                    
                    # Invalide control input at this state
                    Q = self.cf.INF
                        
                self.J[ s ]  = Q
                
                

###############################################################################

class PolicyEvaluatorWithLookUpTable( PolicyEvaluator ):
    """ Evaluate the cost2o of a given control law """
    
    ############################
    def __init__(self, ctl , grid_sys , cost_function , final_time = 0 ):
        
        PolicyEvaluator.__init__(self, ctl , grid_sys, cost_function, final_time)
        
        self.compute_lookuptable()
    
    
    ###############################
    def compute_lookuptable(self):
        """ One step of value iteration """
        
        start_time = time.time()
        print('Computing g(x,u,t) and X-next look-up table..  ', end = '')
        
        self.x_next_table = np.zeros( ( self.grid_sys.nodes_n , self.sys.n ) , dtype = float ) # lookup table for dynamic
        self.G            = np.zeros(   self.grid_sys.nodes_n                , dtype = float ) # lookup table for cost

        # For all state nodes        
        for s in range( self.grid_sys.nodes_n ):  
            
                x = self.grid_sys.state_from_node_id[ s , : ]
                
                ######################################
                # Action taken by the controller
                r = self.ctl.rbar
                u = self.ctl.c( x , r , self.t )   
                ######################################
                
                # Forward dynamics 
                x_next = self.sys.f( x , u , self.t ) * self.grid_sys.dt + x
                
                # Save to llokup table
                self.x_next_table[s,:] = x_next
                
                # If action is in allowable set
                if self.sys.isavalidinput( x , u ):
                    
                    # if the next state is not out-of-bound
                    if self.sys.isavalidstate(x_next):
                        
                        self.G[ s ] = self.cf.g( x , u , self.t ) * self.grid_sys.dt
                    
                    else:
                        # Out of bound cost (J_interpol return 0 in this case)
                        self.G[ s ] = self.cf.INF
                
                else:
                    # Not allowable input at this state
                    self.G[ s ] = self.cf.INF
        
        # Print update
        computation_time = time.time() - start_time
        print('completed in %4.2f sec'%computation_time)
    
                
    ###############################
    def compute_backward_step(self):
        """ One step of value iteration """
        
        self.J       = np.zeros(  self.grid_sys.nodes_n , dtype = float )
        self.Jx_next = np.zeros(  self.grid_sys.nodes_n , dtype = float )
        
        # Computing the J_next of all x_next in the look-up table
        self.Jx_next = self.J_interpol( self.x_next_table )
        
        # Matrix version of computing all Q values
        self.J       = self.G + self.alpha * self.Jx_next


            


'''
#################################################################
##################          Main                         ########
#################################################################
'''


if __name__ == "__main__":     
    """ MAIN TEST """


    from pyro.dynamic  import pendulum
    from pyro.planning import discretizer
    from pyro.analysis import costfunction

    sys  = pendulum.SinglePendulum()

    # Discrete world 
    grid_sys = discretizer.GridDynamicSystem( sys , [101,101] , [3] )

    # Cost Function
    qcf = costfunction.QuadraticCostFunction.from_sys(sys)

    qcf.xbar = np.array([ -3.14 , 0 ]) # target
    qcf.INF  = 300

    # DP algo
    dp = DynamicProgrammingWithLookUpTable(grid_sys, qcf)
    
    dp.solve_bellman_equation( tol = 1.0 )
    
    dp.plot_cost2go()
    dp.plot_cost2go_3D()
    dp.plot_policy()

    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May  4 10:01:30 2023

@author: alex
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic   import cartpole
from pyro.analysis  import costfunction
from pyro.planning  import trajectoryoptimisation
###############################################################################


# sys
sys  = cartpole.CartPole()

sys.xbar[1] = np.pi # Up-right position

#Max/Min torque
sys.u_ub[0] = +20
sys.u_lb[0] = -20


# Cost function
cf  = costfunction.QuadraticCostFunction.from_sys( sys )

cf.Q[0,0] = 1.0
cf.Q[1,1] = 1.0
cf.R[0,0] = 1

planner = trajectoryoptimisation.DirectCollocationTrajectoryOptimisation( 
           sys , 
           dt = 0.2 , 
           grid = 20 ,
           cost_function = cf 
           )


planner.x_start = np.array([0,0,0,0])
planner.x_goal  = np.array([0,np.pi,0,0])

planner.init_dynamic_plot()
planner.maxiter = 500
planner.compute_optimal_trajectory()
# planner.show_solution()
planner.animate_solution()


# -*- coding: utf-8 -*-
"""
Created on Mon Nov 12 20:28:17 2018

@author: Alexandre
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic  import cartpole
from pyro.planning import randomtree
from pyro.planning import trajectoryoptimisation 
###############################################################################


sys  = cartpole.UnderActuatedRotatingCartPole()

sys.u_ub[0] = +50 # Max torque
sys.u_lb[0] = -50 # Min torque

###############################################################################

x_start = np.array([0,-3.14,0,0])
x_goal  = np.array([0,0,0,0])

rrt = randomtree.RRT( sys , x_start )
    
rrt.u_options = [ sys.u_ub , sys.u_lb ]

rrt.goal_radius          = 1.5
rrt.dt                   = 0.1
rrt.max_nodes            = 10000
rrt.max_solution_time    = 3.0
rrt.max_distance_compute = 1000
rrt.dyna_plot            = False

rrt.find_path_to_goal( x_goal )

###############################################################################

planner = trajectoryoptimisation.DirectCollocationTrajectoryOptimisation( sys , 0.1 , 30 )

planner.x_start = x_start
planner.x_goal  = x_goal

planner.set_initial_trajectory_guest( rrt.traj )

planner.maxiter = 500
planner.compute_optimal_trajectory()
planner.show_solution()
planner.animate_solution( is_3d = True )




# -*- coding: utf-8 -*-
"""
Created on Sun May 12 16:34:44 2019

@author: Alexandre
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.control  import robotcontrollers
from pyro.dynamic  import manipulator
###############################################################################

torque_controlled_robot      = manipulator.TwoLinkManipulator()

# Target
q_desired = np.array([0.5,0.5])

# Joint PID

dof = 2

joint_pid      = robotcontrollers.JointPID( dof )
joint_pid.rbar = q_desired
joint_pid.kp   = np.array([25, 5 ])
joint_pid.kd   = np.array([ 5, 2 ])
joint_pid.ki   = np.array([ 8, 4 ])


# Closed-loops

robot_with_joint_pid    = joint_pid + torque_controlled_robot 

# Simulations
tf = 10
robot_with_joint_pid.x0 = np.array([0,0,0,0,0,0])
robot_with_joint_pid.compute_trajectory( tf )
robot_with_joint_pid.plot_trajectory('xu')
robot_with_joint_pid.plot_internal_controller_states()
robot_with_joint_pid.animate_simulation()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar  1 14:08:03 2024

@author: alex
"""

import numpy as np
import matplotlib.pyplot as plt

from pyro.dynamic import pendulum
from pyro.control import controller
from stable_baselines3 import PPO
from stable_baselines3.common.env_util import make_vec_env

sys = pendulum.InvertedPendulum()

# Physical parameters of the prototype
sys.gravity = 9.8
sys.m1 = 0.300
sys.l1 = 0.4
sys.lc1 = 0.4 
sys.I1 = 0.0

max_torque = 3.0

sys.l_domain = 2 * sys.l1  # graphical domain

# Min/max state and control inputs
sys.x_ub = np.array([+2 * np.pi, +12])
sys.x_lb = np.array([-2 * np.pi, -12])
sys.u_ub = np.array([+max_torque])
sys.u_lb = np.array([-max_torque])

# Cost Function
sys.cost_function.xbar = np.array([0, 0])  # target
sys.cost_function.R[0, 0] = 0.0
sys.cost_function.Q[0, 0] = 1.0
sys.cost_function.Q[1, 1] = 0.0

# DP solution
from pyro.planning import discretizer
from pyro.planning import dynamicprogramming

grid_sys = discretizer.GridDynamicSystem(sys, [201, 201], [21])

dp = dynamicprogramming.DynamicProgrammingWithLookUpTable(grid_sys, sys.cost_function)

dp.solve_bellman_equation(tol=0.01)
dp.clean_infeasible_set()
dp.plot_policy()
dp.plot_cost2go(jmax=10)

# Learning
env = sys.convert_to_gymnasium(dt=0.05, render_mode=None)

env.clipping_states = True  # test

model = PPO("MlpPolicy", env, verbose=1)

from pyro.control.reinforcementlearning import stable_baseline3_controller

ppo_ctl = stable_baseline3_controller(model)


ppo_ctl.plot_control_law(sys=sys, n=100)
plt.show()
plt.pause(0.001)

n_time_steps = 250000
batches = 5
env.render_mode = None
for batch in range(batches):
    model.learn(total_timesteps=int(n_time_steps / batches))
    ppo_ctl.plot_control_law(sys=sys, n=100)
    plt.show()
    plt.pause(0.001)

# Animating rl closed-loop
cl_sys = ppo_ctl + sys

cl_sys.x0 = np.array([-3.14, -0.0])
cl_sys.compute_trajectory(tf=10.0, n=10000, solver="euler")
cl_sys.plot_trajectory("xu")
cl_sys.animate_simulation()



# -*- coding: utf-8 -*-
"""
Created on Sun May 12 16:34:44 2019

@author: Alexandre
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic  import manipulator
from pyro.control  import robotcontrollers
###############################################################################


# Model
torque_controlled_robot   = manipulator.ThreeLinkManipulator3D()

# Target
r_desired = np.array([0.5,1,2])

# Effector pid 

model = torque_controlled_robot

effector_pid      = robotcontrollers.EndEffectorPID( model )
effector_pid.rbar = r_desired
effector_pid.kp   = np.array([100, 100 , 100])
effector_pid.kd   = np.array([ 30,  30 ,  30])
effector_pid.ki   = np.array([ 90,  90 ,  90])

# Closed-loops
robot_with_effector_pid = effector_pid + torque_controlled_robot 

# Simulation
robot_with_effector_pid.x0  = np.array([0,0,0,0,0,0,0,0,0])

robot_with_effector_pid.compute_trajectory( 10 )
robot_with_effector_pid.plot_trajectory('x')
robot_with_effector_pid.plot_internal_controller_states()
robot_with_effector_pid.plot_end_effector_trajectory()
robot_with_effector_pid.animate_simulation(time_factor_video=1, is_3d=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May  1 19:51:49 2020

@author: alex
"""
import numpy as np

from gro640_robots import LaserRobot

from pyro.control.robotcontrollers import EndEffectorKinematicController


# Model cinématique du robot
sys = LaserRobot()

# Contrôleur en position de l'effecteur standard
ctl = EndEffectorKinematicController( sys )

# Cible de position pour l'effecteur
ctl.rbar = np.array([0,-1])

# Dynamique en boucle fermée
clsys = ctl + sys

# Configurations de départs
clsys.x0 =  np.array([0,0.5,0]) # crash 
# clsys.x0 =  np.array([0,0.7,0]) # fonctionne

# Simulation
clsys.compute_trajectory( solver = 'odeint'  )
#clsys.compute_trajectory( solver = 'solve_ivt' )

clsys.plot_trajectory('xu')
clsys.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Mon Nov 12 20:28:17 2018

@author: Alexandre
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic  import pendulum
from pyro.control  import nonlinear
from pyro.planning import randomtree

###############################################################################


sys  = pendulum.DoublePendulum()

###############################################################################

x_start = np.array([-3.14,0,0,0])
x_goal  = np.array([0,0,0,0])

planner = randomtree.RRT( sys , x_start )

t = 12
    
planner.u_options = [
        np.array([-t,-t]),
        np.array([-t,+t]),
        np.array([+t,-t]),
        np.array([+t,+t]),
        np.array([ 0,+t]),
         np.array([ 0,-t]),
        np.array([ 0, 0]),
        np.array([+t, 0]),
        np.array([-t, 0])
        ]

planner.goal_radius          = 0.8
planner.dt                   = 0.1
planner.max_nodes            = 12000
planner.max_solution_time    = 8
planner.max_distance_compute = 500

planner.dyna_plot            = False

planner.find_path_to_goal( x_goal )

planner.plot_tree()

###############################################################################

# Controller

ctl  = nonlinear.ComputedTorqueController( sys , planner.traj )

ctl.w0   = 1.0
ctl.zeta = 0.7

# goal
ctl.rbar = np.array([0,0])


# New cl-dynamic
cl_sys = ctl + sys

# Simultation
cl_sys.x0 = x_start
cl_sys.plot_trajectory('xu')
cl_sys.plot_phase_plane_trajectory()
cl_sys.animate_simulation()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May  1 19:51:49 2020

@author: alex
"""
import numpy as np
import matplotlib

from gro640_robots import DrillingRobot
from gro640_robots import DrillingRobotOnJig
from abcd1234      import CustomDrillingController


# Model dynamique du robot
sys = DrillingRobotOnJig()

# Controller
model = DrillingRobot()
ctl   = CustomDrillingController( model ) # Empty do nothing controller template

# Closed-loop dynamic
clsys = ctl + sys

# États initiaux
#clsys.x0 =  np.array([0.7,1.45,-1.4,0,0,0]) #  Tombe dans le trou
clsys.x0 =  np.array([0,1.4,-1.3,0,0,0]) #

# Simulation
tf = 6
n = 100001
clsys.compute_trajectory( tf, n, 'euler')
clsys.plot_trajectory('x')
clsys.plot_trajectory('u')


# Exemple extraction des données pour analyse
t        = clsys.traj.t
q_traj   = clsys.traj.x[:,0:3]  # Trajectoire des angles du robot
dq_traj  = clsys.traj.x[:,3:6]  # Trajectoire des vitesses du robot
tau_traj = clsys.traj.u         # Trajectoire des couples du robot

# Force de contact
##########################################################################################
# NOTE: les impacts de sont pas modèlisés dans la simulation, le comportement 
# transitoire obtenu n'est pas représentatif de ce que serait le vrai comportement.
# Le comportement quasi-statique (lorsque les accélérations sont faibles) est fiable.
##########################################################################################

f_traj   = np.zeros((n,3))
for i in range(n):
    f_traj[i,:] = sys.f_ext( q_traj[i,:] , dq_traj[i,:] )

fig , plots = matplotlib.pyplot.subplots(3, sharex=True, figsize=(4, 3),
                              dpi=300, frameon=True)
plots[0].plot( t , f_traj[:,0] , 'b') 
plots[0].set_ylabel('F_x [N]', fontsize=5 )
plots[0].grid(True)
plots[0].tick_params( labelsize = 5 )
plots[1].plot( t , f_traj[:,1] , 'b') 
plots[1].set_ylabel('F_y [N]', fontsize=5 )
plots[1].grid(True)
plots[1].tick_params( labelsize = 5 )
plots[2].plot( t , f_traj[:,2] , 'b')
plots[2].set_ylabel('F_z [N]', fontsize=5 )
plots[2].grid(True)
plots[2].tick_params( labelsize = 5 )
plots[2].set_xlabel('Time [sec]', fontsize=5 )
fig.canvas.manager.set_window_title('Contact forces')

# End-effector position
clsys.plot_end_effector_trajectory()

# Robot animation
clsys.animate_simulation( is_3d = True )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Oct 16 22:27:47 2022

@author: alex
"""

import numpy as np

from pyro.dynamic  import vehicle_propulsion
from pyro.analysis import costfunction
from pyro.planning import dynamicprogramming 
from pyro.planning import discretizer

sys  = vehicle_propulsion.LongitudinalFrontWheelDriveCarWithWheelSlipInput()

sys.x_ub[0] = 60
sys.x_lb[0] = 0
sys.x_ub[1] = 20
sys.x_lb[1] = 0

# Discrete world 
grid_sys = discretizer.GridDynamicSystem( sys , [101,101] , [3] )

# Cost Function
qcf = costfunction.QuadraticCostFunction.from_sys(sys)

qcf.xbar = np.array([ 0, 0 ]) # target
qcf.INF  = 1000

qcf.Q[0,0] = 0.0
qcf.Q[1,1] = 0.0
qcf.R[0,0] = 0.0

qcf.S[0,0] = 0.0
qcf.S[1,1] = 10000.0


# DP algo
dp = dynamicprogramming.DynamicProgrammingWithLookUpTable( grid_sys, qcf)
#dp = dprog.DynamicProgramming2DRectBivariateSpline(grid_sys, qcf)

dp.solve_bellman_equation( animate_cost2go = True )



# -*- coding: utf-8 -*-
"""
Created on Fri Aug 07 11:51:55 2015

@author: agirard
"""

###############################################################################
import numpy as np
from copy import copy
from scipy.integrate import cumtrapz
###############################################################################


###############################################################################
# Mother cost function class
###############################################################################

class CostFunction():
    """
    Mother class for cost functions of continuous dynamical systems
    ----------------------------------------------
    n : number of states
    m : number of control inputs
    ---------------------------------------
    J = int( g(x,u,t) * dt ) + h( x(T) , T )

    """

    ############################
    def __init__(self):
        self.INF = 1E3
        self.EPS = 1E-3
        
    ###########################################################################
    # The two following functions needs to be implemented by child classes
    ###########################################################################

    #############################
    def h(self, x, t ):
        """ Final cost function """

        raise NotImplementedError

    #############################
    def g(self, x, u, t ):
        """ step cost function """

        raise NotImplementedError
        
    ###########################################################################
    # Method using h and g
    ###########################################################################
        
    #############################
    def trajectory_evaluation(self, traj):
        """
        
        Compute cost and add it to simulation data

        Parameters
        ----------
        traj : instance of `pyro.analysis.Trajectory`

        Returns
        -------
        new_traj : A new instance of the input trajectory, with updated `J` and
        `dJ` fields

        J : array of size ``traj.time_steps`` (number of timesteps in 
        trajectory)
            Cumulative value of cost integral at each time step. The total 
            cost is
            therefore ``J[-1]``.

        dJ : array of size ``traj.time_steps`` (number of timesteps in 
        trajectory)
            Value of cost function evaluated at each point of the tracjectory.
        """

        dJ = np.empty(traj.time_steps)
        for i in range(traj.time_steps):
            x = traj.x[i, :]
            u = traj.u[i, :]
            t = traj.t[i]
            dJ[i] = self.g( x, u, t)

        J = cumtrapz(y=dJ, x=traj.t, initial=0)

        new_traj = copy(traj)
        new_traj.J = J
        new_traj.dJ = dJ

        return new_traj
    
    
###############################################################################
# Basic cost functions
###############################################################################
     
class QuadraticCostFunction( CostFunction ):
    """ 
    Quadratic cost functions of continuous dynamical systems
    ----------------------------------------------
    n : number of states
    m : number of control inputs
    ---------------------------------------
    J = int( g(x,u,t) * dt ) + h( x(T) , T )
    
    g = xQx + uRu 
    h = xSx
    
    """
    
    ############################
    def __init__(self, n, m):
        
        CostFunction.__init__(self)
        
        # dimensions
        self.n = n
        self.m = m
        
        # nominal values
        self.xbar = np.zeros(self.n)
        self.ubar = np.zeros(self.m)

        # Quadratic cost weights
        self.Q = np.diag( np.ones(n)  )
        self.R = np.diag( np.ones(m)  )
        self.S = np.diag( np.zeros(n) )
        
        # Optionnal zone of zero cost if ||x - xbar || < EPS 
        self.ontarget_check = True
        
    
    ############################
    @classmethod
    def from_sys(cls, sys):
        """ From ContinuousDynamicSystem instance """
        
        instance = cls( sys.n , sys.m )
        
        instance.xbar = sys.xbar
        instance.ubar = sys.ubar
        
        return instance
    

    #############################
    def h(self, x , t = 0):
        """ Final cost function with zero value """
        
        # Delta values with respect to nominal values
        dx = x - self.xbar
        
        # Quadratic terminal cost
        J_f = np.dot( dx.T , np.dot(  self.S , dx ) )
                     
        # Set cost to zero if on target
        if self.ontarget_check:
            if ( np.linalg.norm( dx ) < self.EPS ):
                J_f = 0
        
        return J_f
    
    
    #############################
    def g(self, x, u, t):
        """ Quadratic additive cost """
        
        """
        TODO: Add check in init
        # Check dimensions
        if not x.shape[0] == self.Q.shape[0]:
            raise ValueError(
            "Array x of shape %s does not match weights Q with %d components" \
            % (x.shape, self.Q.shape[0])
            )
        if not u.shape[0] == self.R.shape[0]:
            raise ValueError(
            "Array u of shape %s does not match weights R with %d components" \
            % (u.shape, self.R.shape[0])
            )
        if not y.shape[0] == self.V.shape[0]:
            raise ValueError(
            "Array y of shape %s does not match weights V with %d components" \
            % (y.shape, self.V.shape[0])
            )
        """
            
        # Delta values with respect to nominal values
        dx = x - self.xbar
        du = u - self.ubar
        
        dJ = ( np.dot( dx.T , np.dot(  self.Q , dx ) ) +
               np.dot( du.T , np.dot(  self.R , du ) ) )
        
        # Set cost to zero if on target
        if self.ontarget_check:
            if ( np.linalg.norm( dx ) < self.EPS ):
                dJ = 0
        
        return dJ
    
    
#############################################################################
     
class QuadraticCostFunctionVectorized( CostFunction ):
    """ 
    Vectorized: (x, u , t) can be trajectory of time matrices
    
    Quadratic cost functions of continuous dynamical systems
    ----------------------------------------------
    n : number of states
    m : number of control inputs
    ---------------------------------------
    J = int( g(x,u,t) * dt ) + h( x(T) , T )
    
    g = xQx + uRu 
    h = xSx
    
    """
    
    ############################
    def __init__(self, n, m):
        
        CostFunction.__init__(self)
        
        # dimensions
        self.n = n
        self.m = m

        # Quadratic cost weights
        self.Q = np.diag( np.ones(n)  )
        self.R = np.diag( np.ones(m)  )
        self.S = np.diag( np.zeros(n) )
        
        self.is_vectorized = True
        
    
    ############################
    @classmethod
    def from_sys(cls, sys):
        """ From ContinuousDynamicSystem instance """
        
        instance = cls( sys.n , sys.m )
        
        return instance
    

    #############################
    def h(self, x , t = 0):
        """ Final cost function with zero value """
        
        if x.ndim == 1 :
            
            J_f = x.T @ self.S @ x 
            
            
        else:
            
            # Quadratic terminal cost
            J_f = np.diag( x.T @ self.S @ x )
        
        return J_f
    
    
    #############################
    def g(self, x, u, t):
        """ Quadratic additive cost """
        
        if x.ndim == 1 :
            
            dJ = x.T @ self.Q @ x + u.T @ self.R @ u
            
        else:
        
            dJ = np.diag( x.T @ self.Q @ x ) + np.diag( u.T @ self.R @ u )
        
        return dJ

    

##############################################################################

class TimeCostFunction( CostFunction ):
    """ 
    Mother class for cost functions of continuous dynamical systems
    ----------------------------------------------
    n : number of states
    m : number of control inputs
    ---------------------------------------
    J = int( g(x,u,t) * dt ) + h( x(T) , T ) = T
    
    g = 1
    h = 0
    
    """
    
    ############################
    def __init__(self, xbar ):

        CostFunction.__init__(self)
        
        self.xbar = xbar
        
        self.ontarget_check = True
        
    #############################
    def h(self, x , t = 0):
        """ Final cost function with zero value """
        
        return 0
    
    
    #############################
    def g(self, x , u , t = 0 ):
        """ Unity """
        
        """
        if (x.shape[0] != self.xbar.shape[0]):
            raise ValueError("Got x with %d values, but xbar has %d values" %
                             (x.shape[1], self.xbar.shape[0]))
        """

        dJ = 1
        
        if self.ontarget_check:
            dx = x - self.xbar
            if ( np.linalg.norm( dx ) < self.EPS ):
                dJ = 0
                
        return dJ
    


##############################################################################
class QuadraticCostFunctionWithDomainCheck( CostFunction ):
    """ 
    Quadratic cost functions of continuous dynamical systems
    ----------------------------------------------
    n : number of states
    m : number of control inputs
    ---------------------------------------
    J = int( g(x,u,t) * dt ) + h( x(T) , T )
    
    g = xQx + uRu  if x and u are allowable state and actions
    h = 0          if x and u are allowable state and actions
    
    """
    
    ############################
    def __init__(self, n, m, isavalidstate ):
        
        QuadraticCostFunction.__init__(self, n , m )
        
        self.isavalidstate = isavalidstate
    
    ############################
    @classmethod
    def from_sys(cls, sys):
        """ From ContinuousDynamicSystem instance """
        
        instance = cls( sys.n , sys.m , sys.isavalidstate )
        
        instance.xbar = sys.xbar
        instance.ubar = sys.ubar
        
        return instance
    

    #############################
    def h(self, x , t = 0):
        """ Final cost function with zero value """
        
        # Delta values with respect to nominal values
        dx = x - self.xbar
        
        # Quadratic terminal cost
        J_f = np.dot( dx.T , np.dot(  self.S , dx ) )
                
        # Set cost to INF if not an allowable state
        if not self.isavalidstate( x ):
            J_f = self.INF
            
        # Set cost to zero if on target
        if self.ontarget_check:
            if ( np.linalg.norm( dx ) < self.EPS ):
                J_f = 0
        
        return J_f
    
    
    #############################
    def g(self, x, u, t):
        """ Quadratic additive cost """
            
        # Delta values with respect to nominal values
        dx = x - self.xbar
        du = u - self.ubar
        
        dJ = ( np.dot( dx.T , np.dot(  self.Q , dx ) ) +
               np.dot( du.T , np.dot(  self.R , du ) ) )
                
        # Set cost to INF if not an allowable state
        if not self.isavalidstate( x ):
            dJ = self.INF
            
        # Set cost to zero if on target
        if self.ontarget_check:
            if ( np.linalg.norm( dx ) < self.EPS ):
                dJ = 0
        
        return dJ
    
    
    
##############################################################################

class Reachability( CostFunction ):
    
    ############################
    def __init__(self, isavalidestate , xbar = None , isontarget = None ):

        CostFunction.__init__(self)
        
        self.INF = 1E4
        self.EPS = 0.2
        
        self.isavalidestate = isavalidestate
        
        if isontarget == None:
            # default on target test is a quadratic norm check with xbar
            self.isontarget = self.norm_test 
            self.xbar       = xbar
            
        else:
            # Custom function returning bool from state
            self.isontarget = isontarget 
            
        
    #############################
    def norm_test(self, x , t = 0):
        """ Final cost function with zero value """
        
        dx = x - self.xbar
        
        isontarget = np.linalg.norm( dx ) < self.EPS
        
        return isontarget
    
        
    #############################
    def h(self, x , t = 0):
        """ Final cost """
        
        if self.isontarget( x , t ):
            
            J_f = 0 # Finishing in the target set is very good
            
        else:
            
            J_f = self.INF # Finishing not in the target set is very bad
        
        return J_f
    
    
    #############################
    def g(self, x , u , t = 0 ):
        """ Unity """
        
        if self.isavalidestate( x ):
            
            g = 0 
            
        else:
            
            g = self.INF # The system went on of bounds
        
        return g
    
    

'''
#################################################################
##################          Main                         ########
#################################################################
'''


if __name__ == "__main__":     
    """ MAIN TEST """
    
    pass


# -*- coding: utf-8 -*-
"""
Created on Jun 2 2021

@author: Alex
"""

import numpy as np

from pyro.dynamic import tranferfunction
from pyro.dynamic import statespace
from pyro.dynamic import massspringdamper
from pyro.control import linear


# Plant
m = 1
k = 1
b = 0.5

plant = massspringdamper.SingleMass(m,k,b)

#Controller
kp  = 1
ki  = 1
kd  = 1
tau = 0.1

pid = linear.PIDController(kp,ki,kd,tau)

pid.rbar[0] = 2 # set point


closed_loop_sys = pid + plant

closed_loop_sys.compute_trajectory(30)
closed_loop_sys.plot_trajectory('y')
closed_loop_sys.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Fri Aug 07 11:51:55 2015

@author: agirard
"""

import numpy as np

from scipy.integrate import solve_ivp, odeint
from scipy.interpolate import interp1d

##########################################################################
# Trajectory 
##########################################################################

class Trajectory():
    """ Simulation data """

    _dict_keys = ['x', 'u', 't', 'dx', 'y', 'r', 'J', 'dJ']

    def __init__(self, x, u, t, dx, y, r=None, J=None, dJ=None):
        """
        x:  array of dim = ( time-steps , sys.n )
        u:  array of dim = ( time-steps , sys.m )
        t:  array of dim = ( time-steps , 1 )
        y:  array of dim = ( time-steps , sys.p )
        """

        self.x  = x
        self.u  = u
        self.t  = t
        self.dx = dx
        self.y  = y
        self.r  = r
        self.J  = J
        self.dJ = dJ

        self._compute_size()
        
        
    ############################
    def _asdict(self):
        
        return {k: getattr(self, k) for k in self._dict_keys}
    
    
    ############################
    def save2(self, name = 'trajectory' ):
        
        #TODO ne fonctionne pas cet version, voir avec Françis?
        np.savez(name , **self._asdict())
        
        
    ############################
    def save(self, name = 'trajectory.npy' ):
        
        data = np.array( [ self.x , 
                           self.u , 
                           self.t ,
                           self.dx,
                           self.y ,
                           self.r ,
                           self.J , 
                           self.dJ ] )
        
        np.save( name , data )
        
    
    ############################
    @classmethod
    def load(cls, name):
        try:
            # try to load as new format (np.savez)
            # with np.load(name) as data:
            with np.load(name) as data:
                return cls(**data)

        except ValueError:
            # If that fails, try to load as "legacy" numpy object array
            print('\nLoading Legacy Numpy object .npy')
            data = np.load(name, allow_pickle=True)
            return cls(*data)
        
        
    ############################
    def _compute_size(self):
        
        # print(self.t)
        
        self.time_final = self.t.max()
        self.time_steps = self.t.size

        self.n = self.x.shape[1]
        self.m = self.u.shape[1]
        
        self.ubar = np.zeros( self.m )

        # # Check consistency between signals
        # for arr in [self.x, self.y, self.u, self.dx, self.r, self.J, self.dJ]:
            
        #     if arr is not None:
                
        #         if arr.shape[0] != self.time_steps:
        #             raise ValueError("Result arrays must have same length along axis 0")
                

    ############################
    def t2u(self, t ):
        """ get u from time """

        # Find time index
        i = (np.abs(self.t - t)).argmin()

        # Find associated control input
        u = self.u[i,:]
        
        #if t > self.time_final:
        #    u = self.ubar

        return u
    

    ############################
    def t2x(self, t ):
        """ get x from time """

        # Find time index
        i = (np.abs(self.t - t)).argmin()

        # Find associated state
        return self.x[i,:]
    
    ############################
    def copy(self):
        """ copy object """
        
        x  = self.x.copy()
        u  = self.u.copy()
        t  = self.t.copy()
        dx = self.dx.copy()
        y  = self.y.copy()
        
        if self.r is not None:
            r  = self.r.copy()
        else:
            r = None
        
        if self.J is not None:
            J  = self.J.copy()
            dJ = self.dJ.copy()
        else:
            J  = None
            dJ = None

        new_traj = Trajectory(x, u, t, dx, y, r, J, dJ)
        
        return new_traj
    
    
    ###########################
    def generate_interpol_functions(self):
        """ """
        
        # Create interpol functions
        self.inter_t2x   = interp1d(self.t,self.x.T)
        self.inter_t2u   = interp1d(self.t,self.u.T)
        self.inter_t2dx  = interp1d(self.t,self.dx.T)
        self.inter_t2y   = interp1d(self.t,self.y.T)
        
        if self.r is not None:
            self.inter_t2r = interp1d(self.t,self.r.T)
            
        if self.J is not None:
            self.inter_t2J  = interp1d(self.t,self.J)
            self.inter_t2dJ = interp1d(self.t,self.dJ)
    
    
    ############################
    def re_sample(self, n ):
        """ Return new traj with interpolated new time vector """
        
        self.generate_interpol_functions()
        
        ti = self.t[0]
        tf = self.t[-1]
        
        t = np.linspace( ti, tf, n)
        
        x  = np.zeros(( n, self.x.shape[1] ))
        u  = np.zeros(( n, self.u.shape[1] ))
        dx = np.zeros(( n, self.dx.shape[1] ))
        y  = np.zeros(( n, self.y.shape[1] ))
        #r  = np.zeros(( n, self.r.shape[1] ))
        
        for i in range(n):
            
            x[i,:]  = self.inter_t2x(  t[i] )
            u[i,:]  = self.inter_t2u(  t[i] )
            dx[i,:] = self.inter_t2dx( t[i] )
            y[i,:]  = self.inter_t2y(  t[i] )
            #r[i,:]  = self.inter_t2r( t[i] )


        new_traj = Trajectory(x, u, t, dx, y)
        
        return new_traj
    


##########################################################################
# Simulator
##########################################################################

class Simulator:
    """Simulation Class for open-loop ContinuousDynamicalSystem

    Parameters
    -----------
    cds    : Instance of ContinuousDynamicSystem
    tf     : float : final time for simulation
    n      : int   : number of time steps
    solver : {'ode', 'euler'}
        If ode, uses `scipy.integrate.solve_ivp`. `euler` uses built-in
        solver based on the Euler method.
    """

    ############################
    def __init__(
        self, ContinuousDynamicSystem, tf=10, n=10001, solver='solve_ivt' ):

        self.cds    = ContinuousDynamicSystem
        self.t0     = 0
        self.tf     = tf
        self.n      = n
        self.solver = solver
        self.x0     = self.cds.x0
        self.cf     = self.cds.cost_function 

        # Check Initial condition state-vector
        if self.x0.size != self.cds.n:
            raise ValueError(
                "Number of elements in x0 must be equal to number of states"
            )


    ##############################
    def compute(self, **solver_args ):
        """
        Integrate trough time the equation of motion

        Parameters
        -----------
        kwargs: Keyword arguments passed through to the solver (e.g. solve_ivp)
        
        """
        
        ##############################
        if self.solver == 'solve_ivt':
            
            if self.n is not None:
                t_eval = np.linspace(self.t0 , self.tf , int(self.n))
            else:
                t_eval = None
                
            # solve_ivp takes arguments in reverse order of fsim
            def solverfun(t, y):
                dy = self.cds.fsim(y, t)
                return dy

            sol = solve_ivp(
                solverfun,
                t_span    = [self.t0, self.tf],
                y0        = self.x0,
                t_eval    = t_eval,
                **solver_args
            )

            # Compute inputs-output values
            t_sol = sol.t
            n_sol = t_sol.shape[0]
            x_sol = sol.y.transpose()
            y_sol  = np.zeros((n_sol, self.cds.p ))
            u_sol  = np.zeros((n_sol, self.cds.m))
            dx_sol = np.zeros((n_sol, self.cds.n))

            for i in range(n_sol):
                ti = t_sol[i]
                xi = x_sol[i, :]
                ui = self.cds.t2u( ti )

                dx_sol[i,:] = self.cds.f( xi , ui , ti )
                y_sol[i,:]  = self.cds.h( xi , ui , ti )
                u_sol[i,:]  = ui
                
        
        ##############################
        elif self.solver == 'euler':
            
            npts = 10001 if self.n is None else int(self.n)
            
            t_sol  = np.linspace(self.t0 , self.tf , npts)
            x_sol  = np.zeros((npts, self.cds.n))
            dx_sol = np.zeros((npts, self.cds.n))
            u_sol  = np.zeros((npts, self.cds.m))
            y_sol  = np.zeros((npts, self.cds.p))

            # Initial State
            x_sol[0,:] = self.x0
            
            dt = ( self.tf + 0.0 - self.t0 ) / ( npts - 1 )
            
            for i in range(npts):

                ti = t_sol[i]
                xi = x_sol[i,:]
                ui = self.cds.t2u( ti )

                if i+1 < npts:
                    dx_sol[i]    = self.cds.f( xi , ui , ti )
                    x_sol[i+1,:] = dx_sol[i] * dt + xi

                y_sol[i,:] = self.cds.h( xi , ui , ti )
                u_sol[i,:] = ui
                
        ##############################
        elif self.solver == 'odeint':
            
            npts = 10001 if self.n is None else int(self.n)
            
            t_sol = np.linspace(self.t0 , self.tf , npts)
            x_sol = odeint( self.cds.fsim , self.x0 , t_sol)

            # Compute inputs-output values
            y_sol  = np.zeros(( self.n , self.cds.p ))
            u_sol  = np.zeros((self.n,self.cds.m))
            dx_sol = np.zeros((self.n,self.cds.n))

            for i in range(self.n):
                ti = t_sol[i]
                xi = x_sol[i,:]
                ui = self.cds.t2u( ti )

                dx_sol[i,:] = self.cds.f( xi , ui , ti )
                y_sol[i,:]  = self.cds.h( xi , ui , ti )
                u_sol[i,:]  = ui
                
        ##############################
        else :
            
            # self.solver == ???
            print('Check the solver argument: self.solver ==???')
            raise NotImplementedError
            
                
        #########################
        traj = Trajectory(
          x = x_sol,
          u = u_sol,
          t = t_sol,
          dx= dx_sol,
          y = y_sol
        )
        #########################
        
        # Compute Cost function
        if self.cf is not None :
            traj = self.cf.trajectory_evaluation( traj )
        
        return traj



###############################################################################
# Closed Loop Simulator
###############################################################################
    
class CLosedLoopSimulator(Simulator):
    """ 
    Simulation Class for closed-loop ContinuousDynamicalSystem 
    --------------------------------------------------------
    CLSystem  : Instance of ClosedLoopSystem
    tf : final time
    n  : number of point
    solver : 'ode' or 'euler'
    --------------------------------------------------------
    Use this class instead of Simulation() in order to access
    internal control inputs
    """
    
    ############################
    def __init__(self, ClosedLoopSystem, tf=10, n=10001, solver='ode'):
        
        # Mother class init
        Simulator.__init__(self, ClosedLoopSystem, tf, n, solver)
        
        # Special cases
        
        # Use the plant cost function for closed-loop sys
        self.plant_cf = ClosedLoopSystem.plant.cost_function
        

    ###########################################################################
    def compute(self):
        
        traj = Simulator.compute(self)
        
        u = self._compute_control_inputs( traj )

        cl_traj = Trajectory(
            x  = traj.x,
            u  = u,
            t  = traj.t,
            dx = traj.dx,
            y  = traj.y,
            r  = traj.u.copy() # reference is input of global sys
        )
        
        # Compute Cost function
        if self.plant_cf is not None :
            
            cl_traj = self.plant_cf.trajectory_evaluation( cl_traj )

        return cl_traj
        

    ###########################################################################
    def _compute_control_inputs(self, traj):
        """ Compute internal control inputs of the closed-loop system """

        r = traj.u.copy() # reference is input of combined sys
        npts = traj.t.shape[0]
        u = np.zeros([npts, self.cds.plant.m])

        # Compute internal input
        for i in range(npts):

            ri = r[i,:]
            yi = traj.y[i,:]
            ti = traj.t[i]

            ui = self.cds.controller.c( yi , ri , ti )

            u[i,:] = ui

        return u


###############################################################################
# Dynamic Closed Loop Simulator
###############################################################################
    
class DynamicCLosedLoopSimulator( CLosedLoopSimulator ):
    """ 
    Specific simulator for extracting internal control signal
    """

    ###########################################################################
    def _compute_control_inputs(self, traj ):
        """ Compute internal control inputs of the closed-loop system """

        r = traj.u.copy() # reference is input of combined sys
        npts = traj.t.shape[0]
        u = np.zeros([npts ,self.cds.plant.m])

        # Compute internal input signal_proc
        for i in range(npts):

            ri = r[i,:]
            yi = traj.y[i,:]
            xi = traj.x[i,:]
            ti = traj.t[i]

            # extract internal controller states
            xi,zi = self.cds._split_states( xi ) 

            ui = self.cds.controller.c( zi, yi , ri , ti )

            u[i,:] = ui

        return u



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun  2 11:17:40 2021

@author: alex
"""


import numpy as np
import scipy.signal as signal
import matplotlib
import matplotlib.pyplot as plt

from pyro.dynamic import ContinuousDynamicSystem
from pyro.dynamic import linearize

# Embed font type in PDF
matplotlib.rcParams['pdf.fonttype'] = 42
matplotlib.rcParams['ps.fonttype']  = 42

###############################################################################
class TransferFunction( ContinuousDynamicSystem ):
    """Time-invariant transfer function representation of a dynamic system

    Y(s) / U(s) =   [ num ] / [den]

    """
    ############################################
    def __init__(self, num, den):
        
        self.num = num
        self.den = den

        n = den.shape[0] - 1
        
        ContinuousDynamicSystem.__init__( self, n, 1, 1)
        
        self.compute_ss()
        self.compute_poleszeros()
        
        # Plot params
        self.figsize = (5,3)
        self.dpi = 300
        self.fontsize = 5
        
    #############################################
    def compute_ss(self):

        A, B, C, D = signal.tf2ss(self.num, self.den)

        self.A = A
        self.B = B
        self.C = C
        self.D = D
        
    #############################################
    def compute_poleszeros(self):
        
        tf = signal.TransferFunction(self.num, self.den)
        
        self.poles = tf.poles
        self.zeros = tf.zeros
        
    #############################################
    def f(self, x, u, t):

        dx = np.dot(self.A, x) + np.dot(self.B, u)

        return dx
    
    #############################################
    def h(self, x, u, t):
        
        y = np.dot(self.C, x) + np.dot(self.D, u)
        
        return y
    
    ############################################
    def bode_plot(self):
        """ Plot frequency response """
        
        tf = signal.TransferFunction(self.num, self.den)
        
        w, mag, phase = tf.bode()
        
        fig , plots = plt.subplots(2, sharex=True, figsize=self.figsize, 
                                  dpi=self.dpi, frameon=True)
        
        plots[0].semilogx(w, mag)
        plots[1].semilogx(w, phase)
        
        plots[0].set_ylabel(self.output_label[0] + ' ' + self.output_units[0] +'\n-------------------\n'+ self.input_label[0] + ' ' + self.input_units[0]
                 , fontsize= self.fontsize )
        plots[1].set_ylabel( 'Phase [rad]', fontsize= self.fontsize )
        plots[1].set_xlabel( 'Freq [rad/sec]', fontsize= self.fontsize )
        
        for i in [0,1]:
            plots[i].grid(True)
            plots[i].tick_params( labelsize = self.fontsize )
        
        fig.canvas.manager.set_window_title('Bode plot of ' + self.name)
        
        plt.show()
        
    ############################################
    def pz_map(self):
        """ Plot poles and zeros """
        
        self.compute_poleszeros()
        
        fig , plot = plt.subplots(1, sharex=True, figsize=self.figsize, 
                                  dpi=self.dpi, frameon=True)
        
        plot.plot( self.poles.real, self.poles.imag, 'x')
        plot.plot( self.zeros.real, self.zeros.imag, 'o')
        plot.grid(True)
        plot.tick_params( labelsize = self.fontsize )
        
        fig.canvas.manager.set_window_title('Poles and zeros of ' + self.name)
        
        plt.show()
        




#################################################################
def ss2tf( ss, u_index , y_index):
    """
    Compute the transfer function of a given input and output of a state-space
    """
    nums, den = signal.ss2tf(ss.A, ss.B, ss.C, ss.D, u_index)
    
    num = nums[y_index]
    
    tf = TransferFunction(num, den)
    
    tf.name = (ss.output_label[y_index] + '/' + ss.input_label[u_index] + 
               ' transfer function of ' + ss.name )

    tf.output_label[0] = ss.output_label[y_index]
    tf.output_units[0] = ss.output_units[y_index]
    tf.input_label[0]  = ss.input_label[u_index]
    tf.input_units[0]  = ss.input_units[u_index]
    
    return tf
        

'''
#################################################################
##################          Main                         ########
#################################################################
'''


if __name__ == "__main__":     
    """ MAIN TEST """
    
    num = np.array([1])
    den = np.array([1,0,1])
    
    #TF = signal.TransferFunction( num, den)
    
    sys = TransferFunction( num , den)
    
    #sys.bode_plot()
    
    from pyro.dynamic import pendulum
    
    non_linear_sys = pendulum.SinglePendulum()
    
    non_linear_sys.lc1     = 1
    non_linear_sys.m1      = 2
    non_linear_sys.I1      = 2
    non_linear_sys.d1      = 2
    non_linear_sys.gravity = 9.81
    
    linearized_sys = linearize( non_linear_sys )
    siso_sys       = ss2tf( linearized_sys, 0, 0)
    
    print('Poles',siso_sys.poles)
    print('num',siso_sys.num)
    print('den',siso_sys.den)
    
    siso_sys.bode_plot()
    
    
    


# -*- coding: utf-8 -*-
"""
Created on Sun May 12 16:34:44 2019

@author: Alexandre
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.planning import randomtree
from pyro.dynamic  import manipulator
###############################################################################

torque_controlled_robot = manipulator.FiveLinkPlanarManipulatorwithObstacles()
speed_controlled_robot  = manipulator.SpeedControlledManipulator.from_manipulator( torque_controlled_robot )

q_start = np.array([0,0,0,0,0])
q_goal  = np.array([1.57,0,0,0,0])

planner = randomtree.RRT( speed_controlled_robot , q_start )

t = 0.5
    
planner.u_options = [
        np.array([ +t, -0, +0, +0, +0]),
        np.array([ -t, -0, +0, +0, +0]),
        np.array([ -t, -t, +t, +t, +t]),
        np.array([ +t, +t, -t, -t, -t]),
        np.array([ -t, -t, -t, +t, +t]),
        np.array([ +t, +t, +t, -t, -t])
        ]

planner.goal_radius = 0.4
planner.max_nodes   = 20000
planner.dyna_plot   = False
planner.max_solution_time = 20


planner.find_path_to_goal( q_goal )

#planner.save_solution('fivelinkplan.npy')
planner.plot_tree()
planner.show_solution()
planner.animate_solution()



# -*- coding: utf-8 -*-
"""
Created on Tue Nov 13 11:05:07 2018

@author: Alexandre
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic import vehicle_steering
from pyro.planning import randomtree
###############################################################################

sys  = vehicle_steering.KinematicBicyleModel()

###############################################################################

# Planning
planner = randomtree.RRT( sys )

# Set domain
sys.x_ub = np.array([+10,+10,+6.28])
sys.x_lb = np.array([-10,-10,-6.284])

# Planning Problem
planner.x_start = np.array([0,0,0])
planner.x_goal  = np.array([5,5,0])


# Control actions
speed    = 1
steering = 0.5

planner.u_options = [
        np.array([ speed,-steering]),
        np.array([ speed,+steering]),
        np.array([ speed,0]),
        np.array([ -speed,0])
        ]

planner.goal_radius = 1.0
planner.dt          = 0.1
planner.steps       = 5

planner.compute_solution()

planner.plot_tree()
planner.plot_tree_3d()
planner.show_solution()
planner.animate_solution()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Oct 16 22:27:47 2022

@author: alex
"""

import numpy as np

from pyro.dynamic  import suspension
from pyro.analysis import costfunction
from pyro.planning import dynamicprogramming 
from pyro.planning import discretizer

sys  = suspension.QuarterCarOnRoughTerrain()

sys.mass = 0.5
sys.b    = 0.5
sys.k    = 8.0
sys.vx   = 10.0

# Set domain
sys.x_ub = np.array([+12, +1, +40])
sys.x_lb = np.array([-12, -1, +0])

sys.u_ub = np.array([+40])
sys.u_lb = np.array([-40])

# Discrete world
grid_sys = discretizer.GridDynamicSystem(sys, (51, 51, 101), [11], 0.05)

# Cost Function
qcf = costfunction.QuadraticCostFunction.from_sys(sys)

qcf.xbar = np.array([ 0.0 , 0.0, 20 ]) # target
qcf.INF  = 100000
qcf.EPS  = 0.5

qcf.Q[0,0] = 2.0
qcf.Q[1,1] = 5.0
qcf.Q[2,2] = 0.0

qcf.R[0,0] = 0.1

qcf.S[0,0] = 0.0
qcf.S[1,1] = 0.0
qcf.S[2,2] = 0.0

# DP algo
dp = dynamicprogramming.DynamicProgrammingWithLookUpTable( grid_sys, qcf)

dp.alpha = 0.99
dp.solve_bellman_equation()


ctl = dp.get_lookup_table_controller()

cl_sys = ctl + sys

# Simulation and animation
cl_sys.x0   = np.array([0,0,-60])
cl_sys.compute_trajectory( 10, 10001, 'euler')
cl_sys.plot_trajectory('xu')
cl_sys.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Fri Nov 16 12:05:08 2018

@author: Alexandre
"""
###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic  import pendulum
from pyro.control  import nonlinear
from pyro.analysis import simulation
###############################################################################

sys  = pendulum.DoublePendulum()

#traj = simulation.Trajectory.load('double_pendulum_rrt.npy')
#traj = simulation.Trajectory.load('double_pendulum_directcollocation.npy')
traj = simulation.Trajectory.load('double_pendulum_directcollocation_hires.npy')

# Controller
ctl  = nonlinear.ComputedTorqueController( sys , traj )

# goal
ctl.rbar = np.array([0,0])
ctl.w0   = 5
ctl.zeta = 1

# New cl-dynamic
cl_sys = ctl + sys

# Simultation
cl_sys.x0 = np.array([-3.14,0,0,0])
cl_sys.plot_trajectory('xu')
cl_sys.animate_simulation()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 23 20:24:14 2020

@author: alex
"""

# -*- coding: utf-8 -*-
"""
Created on Fri Nov 16 12:05:08 2018

@author: Alexandre
"""
###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic import pendulum
from pyro.control import nonlinear
###############################################################################

sys = pendulum.DoublePendulum()
ctl  = nonlinear.ComputedTorqueController( sys )

ctl.w0   = 1.5
ctl.zeta = 0.5 


# Custom time-based reference for controller
def t2r( t ):
    r = np.array([0.,0.])
    r[1] = 1.0 * np.sin( 5 * t )
    return r

ctl.t2r = t2r

# New cl-dynamic
cl_sys = ctl + sys

# Simultation
cl_sys.x0  = np.array([-3.14,1,0,0])
cl_sys.compute_trajectory( tf = 20 )
cl_sys.plot_trajectory()
cl_sys.plot_phase_plane_trajectory(0, 2)
cl_sys.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Mon Nov 12 20:28:17 2018

@author: Alexandre
"""
##############################################################################
import numpy as np
##############################################################################
from pyro.dynamic  import pendulum
from pyro.planning import OpenLoopController
##############################################################################

# Dynamic system
sys  = pendulum.SinglePendulum()

# Openloop controller loading RRT solution
ctl = OpenLoopController.load_from_file('pendulum_rrt.npy')

# Closing the loop
cl_sys = ctl + sys

# Simulation
cl_sys.x0 = np.array([0.1,0])
cl_sys.plot_trajectory('xu')
cl_sys.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Fri Nov 16 10:24:51 2018

@author: Alexandre
"""

import numpy as np
############################################################################
from pyro.dynamic  import integrator
from pyro.control  import linear
############################################################################

# Double integrator
si = integrator.SimpleIntegrator()
di = integrator.DoubleIntegrator()
ti = integrator.TripleIntegrator()

# Controller 
ctl      = linear.ProportionalController()

ctl.K[0,0] = 1.0

# New cl-dynamic
clsi = ctl + si

clsi.x0 = np.array([10])
clsi.plot_trajectory('xu')
clsi.plot_phase_plane_trajectory(0, 0)

cldi = ctl + di

cldi.x0 = np.array([10, 0])
cldi.plot_trajectory('xu')
cldi.plot_phase_plane_trajectory(0, 1)

clti = ctl + ti

clti.x0 = np.array([0.1,0,0])
clti.plot_trajectory('xu')
clti.plot_phase_plane_trajectory_3d()


# -*- coding: utf-8 -*-
"""
Created on 19/11/2019

@author: Pierre
"""
###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic import pendulum
from adaptive_computed_torque import SinglePendulumAdaptativeController
###############################################################################

sys = pendulum.SinglePendulum()


sys.m1       = 3
sys.l1       = 2
sys.lc1      = 1
sys.I1       = 0
sys.gravity  = 9.81 


# Controller

ctl = SinglePendulumAdaptativeController(sys)

# Sys ID initial guess
m1_hat = 1
l1_hat = 1

ctl.z0[0] = m1_hat * l1_hat ** 2
ctl.z0[1] = m1_hat * l1_hat * 9.81

# Controller Param

ctl.K      = 1
ctl.lam    = 1
ctl.P[0,0] = 10
ctl.P[1,1] = 10


# Set Point
q_target = np.array([3.14])
ctl.rbar = q_target

# New cl-dynamic
cl_sys = ctl + sys

cl_sys.state_label[2] = 'H'
cl_sys.state_label[3] = 'mgl'


# Simultation
cl_sys.x0[0]  = 0

cl_sys.compute_trajectory(tf=20, n=20001) # solver='euler')
cl_sys.plot_phase_plane_trajectory()
cl_sys.plot_trajectory_with_internal_states()
cl_sys.animate_simulation()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun May 24 11:51:18 2020

@author: alex
"""

import numpy as np

from pyro.dynamic.pendulum      import DoublePendulum
from pyro.analysis.costfunction import QuadraticCostFunction
from pyro.dynamic.statespace    import linearize
from pyro.control.lqr           import synthesize_lqr_controller


# Non-linear model
sys = DoublePendulum()
    
# Linear model
ss  = linearize( sys , 0.01 )

# Cost function
cf  = QuadraticCostFunction.from_sys( sys )
cf.R[0,0] = 1000
cf.R[1,1] = 10000

# LQR controller
ctl = synthesize_lqr_controller( ss , cf )

# Simulation Closed-Loop Non-linear with LQR controller
cl_sys = ctl + sys
cl_sys.x0 = np.array([0.4,0,0,0])
cl_sys.compute_trajectory()
cl_sys.plot_trajectory('xu')
cl_sys.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Fri Nov 16 12:01:07 2018

@author: Alexandre
"""

###############################################################################
import numpy as np
import matplotlib.pyplot as plt
###############################################################################
from pyro.dynamic import vehicle_steering
###############################################################################
import advanced_vehicles
import test_vehicle_controllers
###############################################################################

# "Fake" controller - Varying inputs (delta, T_f, T_r) throughout time (change in linear.py)
ctl = test_vehicle_controllers.kinematicInputs()
# Vehicule dynamical system
sys = vehicle_steering.KinematicBicyleModel()

# Set default wheel velocity and steering angle
cl_sys = ctl+sys


# Plot open-loop behavior (ex: np.array[intial_conditions], time_of_simulation)
cl_sys.x0 = np.array([0,0,0])
cl_sys.plot_trajectory()

# Rebuild x,u and t from simulation
x = cl_sys.traj.x
u = cl_sys.traj.u
t = cl_sys.traj.t   

# Plot the vehicle's trajectory
figsize   = (7, 4)
dpi       = 100
plt.figure(2,figsize=figsize, dpi=dpi)
plt.axes(xlim=(-20,150), ylim=(-5,200))  
plt.plot(x[:,0],x[:,1],'--g', label = 'Kinematic Model')     
plt.legend(fontsize ='15')   
plt.legend(fontsize ='15')
plt.title("Vehicle's cartesian position", fontsize=20)
plt.xlabel('X position (m)')
plt.ylabel('Y position (m)')
plt.show()

# Animate the simulation
cl_sys.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Jun 2 2021

@author: Alex
"""

from pyro.dynamic import massspringdamper

sys = massspringdamper.TwoMass()

sys.m1 = 2
sys.m2 = 3
sys.k1 = 5
sys.k2 = 2
sys.b1 = 0.1
sys.b2 = 0.2

sys.compute_ABCD()


sys.plot_linearized_bode()
sys.plot_linearized_pz_map()

sys.x0[1] = 1

sys.plot_trajectory()
sys.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Fri Nov 16 12:05:08 2018

@author: Alexandre
"""
###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic  import pendulum
from pyro.planning import plan
###############################################################################

sys  = pendulum.DoublePendulum()

ctl  = plan.OpenLoopController.load_from_file( 'double_pendulum_directcollocation_hires.npy')

# New cl-dynamic
cl_sys = ctl + sys

# Simultation
cl_sys.x0 = np.array([-3.14,0,0,0])
cl_sys.compute_trajectory(10)
cl_sys.plot_phase_plane_trajectory(0,2)
cl_sys.plot_trajectory('xu')
cl_sys.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Wed May  8 12:43:21 2019

@author: alxgr
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.control import controller
###############################################################################


###############################################################################
# Mother Class
###############################################################################

class RobotController( controller.StaticController ) :
    """   
    Mother class for robot controllers
    ---------------------------------------
    r  : reference signal vector       k x 1
    y  : sensor signal vector          p x 1
    u  : control inputs vector         m x 1
    t  : time                          1 x 1
    ---------------------------------------
    u = c( y , r , t )
    """
    
    ############################
    def __init__(self, dof = 1):
        """ """

        # Dimensions
        self.k = dof      # ref signal dimension
        self.m = dof      # assuming fully-actuated system
        self.p = dof * 2  # y = x ( full state feedback )
        
        controller.StaticController.__init__( self, self.k, self.m, self.p)
        
        # Label
        self.name = 'Robot Controller'
        
        # DoF
        self.dof = dof
        
    #############################
    def x2q( self, x ):
        """ from state vector (x) to angle and speeds (q,dq) """
        
        q  = x[ 0        :     self.dof   ]
        dq = x[ self.dof : 2 * self.dof   ]
        
        return [ q , dq ]
        


###############################################################################
# Joint PD
###############################################################################
        
class JointPD( RobotController ) :
    """ 
    Linear controller for mechanical system with full state feedback (y=x)
    Independent PID for each DOF
    ---------------------------------------
    r  : reference signal vector  dof x 1
    y  : sensor signal vector     n   x 1
    u  : control inputs vector    dof x 1
    t  : time                     1   x 1
    ---------------------------------------
    u = c( y , r , t ) = (r - q) * kp + ( dq ) * kd 

    """
    
    ############################
    def __init__(self, dof = 1, kp = 1, kd = 0):
        """ """
        
        RobotController.__init__( self, dof )
        
        # Label
        self.name = 'Joint PD Controller'
        
        # Gains
        self.kp = np.ones( dof  ) * kp
        self.kd = np.ones( dof  ) * kd
        
    
    #############################
    def c( self , y , r , t = 0 ):
        """ 
        Feedback static computation u = c(y,r,t)
        
        INPUTS
        y  : sensor signal vector     p x 1
        r  : reference signal vector  k x 1
        t  : time                     1 x 1
        
        OUTPUTS
        u  : control inputs vector    m x 1
        
        """
        
        u = np.zeros(self.m) 
        
        # Ref
        q_d = r
        
        # Feedback from sensors
        x = y
        [ q , dq ] = self.x2q( x )
        
        # Error
        e  = q_d - q
        de =     - dq
        
        # PIDs
        u = e * self.kp + de * self.kd 
        
        return u
    


###############################################################################
# End-Effector PD
###############################################################################
        
class EndEffectorPD( RobotController ) :
    """ 
    PID in effector coordinates, using the Jacobian of the system
    ---------------------------------------
    r  : reference signal vector  e   x 1
    y  : sensor signal vector     n   x 1
    u  : control inputs vector    dof x 1
    t  : time                     1   x 1
    ---------------------------------------
    u = c( y , r , t ) = (r - q) * kp + ( dq ) * kd 

    """
    
    ############################
    def __init__(self, robot, kp = 1, ki = 0, kd = 0):
        """ """
        
        # Using from model
        self.fwd_kin = robot.forward_kinematic_effector
        self.J       = robot.J
        self.e       = robot.e # nb of effector dof
        
        RobotController.__init__( self , robot.dof )
        
        # Label
        self.name = 'End-Effector PID Controller'
        
        # Gains
        self.kp = np.ones( self.dof  ) * kp
        self.kd = np.ones( self.dof  ) * kd
        
    
    #############################
    def c( self , y , r , t = 0 ):
        """ 
        Feedback static computation u = c(y,r,t)
        
        INPUTS
        y  : sensor signal vector     p x 1
        r  : reference signal vector  k x 1
        t  : time                     1 x 1
        
        OUTPUTS
        u  : control inputs vector    m x 1
        
        """
        
        u = np.zeros(self.m) 
        
        # Feedback from sensors
        x = y
        [ q , dq ] = self.x2q( x )
        
        # Jacobian computation
        J = self.J( q )
        
        # Ref
        r_desired   = r
        r_actual    = self.fwd_kin( q )
        dr          = np.dot( J , dq )
        
        # Error
        e  = r_desired - r_actual
        de =           - dr
        
        # Effector space PID
        f = e * self.kp + de * self.kd 
        
        # From effector force to joint torques
        u = np.dot( J.T , f )
        
        return u
    
    
###############################################################################
# Kinematic Controllers
###############################################################################
        
class EndEffectorKinematicController( RobotController ) :
    """ 
    Kinematic effector coordinates controller using the Jacobian of the system
    ------------------------------------------
    r = r_d : reference signal vector  e   x 1
    y = q   : sensor signal vector     dof x 1
    u = dq  : control inputs vector    dof x 1
    t       : time                     1   x 1
    -------------------------------------------
    u = c( y , r , t ) = J(q)^T *  [ (r - r_robot(q)) * k ]

    """
    
    ############################
    def __init__(self, robot, k = 1 ):
        """ """
        
        # Using functions from robot model
        self.fwd_kin = robot.forward_kinematic_effector
        self.J       = robot.J
        self.e       = robot.e # nb of effector dof
        
        # Dimensions
        self.dof = robot.dof
        self.k   = self.e 
        self.m   = self.dof
        self.p   = self.dof
        
        controller.StaticController.__init__(self, self.k, self.m, self.p)

        # Label
        self.name = 'End Effector Kinematic Controller'
        
        # Gains
        self.gains = np.ones( self.e  ) * k
        
    
    #############################
    def c( self , y , r , t = 0 ):
        """ 
        Feedback static computation u = c(y,r,t)
        
        INPUTS
        y  : sensor signal vector     p x 1
        r  : reference signal vector  k x 1
        t  : time                     1 x 1
        
        OUTPUTS
        u  : control inputs vector    m x 1
        
        """
        
        #u = np.zeros(self.m) 
        
        # Feedback from sensors
        q = y
        
        # Jacobian computation
        J = self.J( q )
        
        # Ref
        r_desired   = r
        r_actual    = self.fwd_kin( q )
        
        # Error
        e  = r_desired - r_actual
        
        # Effector space speed
        dr_r = e * self.gains
        
        # From effector speed to joint speed
        if self.dof == self.e:
            
            dq_r = np.dot( np.linalg.inv( J ) , dr_r )
            
        elif self.dof > self.e:
            
            # Pseudo-inverse of Jacobian
            J_pinv = np.linalg.pinv( J )
            
            dq_r   = np.dot( J_pinv , dr_r )
            
        else:
            
            #TODO
            pass
        
        return dq_r
    
###############################################################################
# End-Effector Kinematic with nullspace
###############################################################################

class EndEffectorKinematicControllerWithNullSpaceTask( EndEffectorKinematicController ) :
    """ 
    Kinematic effector coordinates controller using the Jacobian of the system
    inlcuding secondary control loop v = (q_d - q) * k2 with joint space 
    position target projected on the nullspace
    ------------------------------------------
    r = r_d : reference signal vector  e   x 1
    y = q   : sensor signal vector     dof x 1
    u = dq  : control inputs vector    dof x 1
    t       : time                     1   x 1
    -------------------------------------------
    u = c( y , r , t ) = J# [ (r - r_robot(q)) * k ] + [I -J#J] ((q_d - q) * k2 )

    """
    
    ############################
    def __init__(self, robot, k = 1 , k_null = 1):
        """ """
        
        EndEffectorKinematicController.__init__( self, robot , k )
        
        
        self.gains_null = np.ones( self.dof  ) * k_null
        self.q_d        = np.zeros( self.dof  )
        
    
    #############################
    def c( self , y , r , t = 0 ):
        """ 
        Feedback static computation u = c(y,r,t)
        
        INPUTS
        y  : sensor signal vector     p x 1
        r  : reference signal vector  k x 1
        t  : time                     1 x 1
        
        OUTPUTS
        u  : control inputs vector    m x 1
        
        """
        
        #u = np.zeros(self.m) 
        
        # Feedback from sensors
        q = y
        
        # Jacobian computation
        J = self.J( q )
        
        # Ref
        r_desired   = r
        r_actual    = self.fwd_kin( q )
        
        # Error
        e   = r_desired - r_actual
        q_e = self.q_d  - q    
        
        # Effector space speed 
        dr_r    =   e * self.gains
        dq_null = q_e * self.gains_null
        
        # From effector velocity to joint velocities
        if self.dof == self.e:
            
            dq_r = np.dot( np.linalg.inv( J ) , dr_r )
            
        elif self.dof > self.e:
            
            # Pseudo Inverse
            J_pinv = np.linalg.pinv( J )
            
            # Nullspacr projection Matrix
            Null_p = np.identity( self.dof ) - np.dot(J_pinv,J)
            
            dq_r   = np.dot( J_pinv , dr_r ) + np.dot( Null_p , dq_null )
            
        else:
            
            #TODO
            pass
        
        return dq_r
    


###############################################################################
# Dynamic PID controllers
###############################################################################
        
class JointPID( controller.DynamicController ) :
    """ 
    Linear controller for mechanical system with full state feedback (y=x)
    Independent PID for each DOF
    ---------------------------------------
    r  : reference signal vector  dof x 1
    y  : sensor signal vector     n   x 1
    u  : control inputs vector    dof x 1
    t  : time                     1   x 1
    ---------------------------------------
    u = c( y , r , t ) = (r - q) * kp + ( dq ) * kd + int(r - q) * ki

    """
    
    ############################
    def __init__(self, dof = 1, kp = 1, ki = 0, kd = 0):
        """ """

        # Gains
        self.kp = np.ones( dof  ) * kp
        self.kd = np.ones( dof  ) * kd
        self.ki = np.ones( dof  ) * ki
        
        # Dimensions
        self.k = dof      # ref signal dimension
        self.l = dof      # internal states dimensions
        self.m = dof      # assuming fully-actuated system
        self.p = dof * 2  # y = x ( full state feedback )
        
        controller.DynamicController.__init__( self, self.k, self.l, self.m, self.p)
        
        # Label
        self.name = 'Joint PID Controller'
        
        for i in range(self.l):
            self.internal_state_label[i] = 'Int_e dof ' +str(i)
        
        # DoF
        self.dof = dof
        
        
    #############################
    def x2q( self, x ):
        """ from state vector (x) to angle and speeds (q,dq) """
        
        q  = x[ 0        :     self.dof   ]
        dq = x[ self.dof : 2 * self.dof   ]
        
        return [ q , dq ]
        
    
    #############################
    def c( self , z , y , r , t = 0 ):
        """ 
        Feedback static computation u = c(y,r,t)
        
        INPUTS
        y  : sensor signal vector     p x 1
        r  : reference signal vector  k x 1
        t  : time                     1 x 1
        
        OUTPUTS
        u  : control inputs vector    m x 1
        
        """
        
        u = np.zeros(self.m) 
        
        # Ref
        q_d = r
        
        # Feedback from sensors
        x = y
        [ q , dq ] = self.x2q( x )
        
        # Error
        e  = q_d - q
        de =     - dq
        ie = z          # Integral of errors are internal states 
        
        # PIDs
        u = e * self.kp + de * self.kd + ie * self.ki
        
        return u
    
    
    #############################
    def b(self, z, y, r, t):
        """ 
        INTERNAL CONTROLLER DYNAMIC
        dz/dt = b( z, y, r, t)
        
        INPUTS
        z  : internal states               l x 1
        y  : sensor signal vector          p x 1
        r  : reference signal vector       k x 1
        t  : time                          1 x 1
        
        OUTPUTS
        d z / dt  : time derivative of internal states        l x 1
        """
        
        dz = np.zeros( self.l )
        
        # Ref
        q_d = r
        
        # Feedback from sensors
        x = y
        [ q , dq ] = self.x2q( x )
        
        # Error
        e  = q_d - q
        
        # Derivative of internal state is the error
        dz = e
        
        return dz
    


###############################################################################
# End-Effector PID
###############################################################################
        
class EndEffectorPID( controller.DynamicController ) :
    """ 
    Linear controller for mechanical system with full state feedback (y=x)
    Independent PID for each DOF
    ---------------------------------------
    r  : reference signal vector  dof x 1
    y  : sensor signal vector     n   x 1
    u  : control inputs vector    dof x 1
    t  : time                     1   x 1
    ---------------------------------------
    u = c( y , r , t ) = (rd - r) * kp + ( -dr ) * kd + int(rd- r) * ki

    """
    
    ############################
    def __init__(self, robot, kp = 1, ki = 0, kd = 0):
        """ """
        
        # Using from model
        self.fwd_kin = robot.forward_kinematic_effector
        self.J       = robot.J
        self.e       = robot.e # nb of effector dof
        
        dof = robot.dof

        # Gains
        self.kp = np.ones( dof  ) * kp
        self.kd = np.ones( dof  ) * kd
        self.ki = np.ones( dof  ) * ki
        
        # Dimensions
        self.k = dof      # ref signal dimension
        self.l = dof      # internal states dimensions
        self.m = dof      # assuming fully-actuated system
        self.p = dof * 2  # y = x ( full state feedback )
        
        controller.DynamicController.__init__( self, self.k, self.l, self.m, self.p)
        
        # Label
        self.name = 'End-Effector PID Controller'
        
        for i in range(self.l):
            self.internal_state_label[i] = 'Int_e axis ' +str(i)
        
        # DoF
        self.dof = dof
        
        
    #############################
    def x2q( self, x ):
        """ from state vector (x) to angle and speeds (q,dq) """
        
        q  = x[ 0        :     self.dof   ]
        dq = x[ self.dof : 2 * self.dof   ]
        
        return [ q , dq ]
        
    
    #############################
    def c( self , z , y , r , t = 0 ):
        """ 
        Feedback static computation u = c(y,r,t)
        
        INPUTS
        y  : sensor signal vector     p x 1
        r  : reference signal vector  k x 1
        t  : time                     1 x 1
        
        OUTPUTS
        u  : control inputs vector    m x 1
        
        """
        
        u = np.zeros(self.m) 
        
        # Feedback from sensors
        x = y
        [ q , dq ] = self.x2q( x )
        
        # Jacobian computation
        J = self.J( q )
        
        # Ref
        r_desired   = r
        r_actual    = self.fwd_kin( q )
        dr          = np.dot( J , dq )
        
        # Error
        e  = r_desired - r_actual
        de =           - dr
        ie = z
        
        # Effector space PID
        f = e * self.kp + de * self.kd + ie * self.ki
        
        # From effector force to joint torques
        u = np.dot( J.T , f )
        
        return u
    
    
    #############################
    def b( self, z, y, r, t):
        """ 
        INTERNAL CONTROLLER DYNAMIC
        dz/dt = b( z, y, r, t)
        
        INPUTS
        z  : internal states               l x 1
        y  : sensor signal vector          p x 1
        r  : reference signal vector       k x 1
        t  : time                          1 x 1
        
        OUTPUTS
        d z / dt  : time derivative of internal states        l x 1
        """
        
        dz = np.zeros( self.l )
        
        # Feedback from sensors
        x = y
        [ q , dq ] = self.x2q( x )
        
        # Ref
        r_desired   = r
        r_actual    = self.fwd_kin( q )
        
        # Error
        e  = r_desired - r_actual
        
        # Derivative of internal state is the error
        dz = e
        
        return dz
    

###############################################################################
# Backup
###############################################################################
        
class old_JointPID( RobotController ) :
    """ 
    Linear controller for mechanical system with full state feedback (y=x)
    Independent PID for each DOF
    ---------------------------------------
    r  : reference signal vector  dof x 1
    y  : sensor signal vector     n   x 1
    u  : control inputs vector    dof x 1
    t  : time                     1   x 1
    ---------------------------------------
    u = c( y , r , t ) = (r - q) * kp + ( dq ) * kd + int(r - q) * ki

    """
    
    ############################
    def __init__(self, dof = 1, kp = 1, ki = 0, kd = 0):
        """ """
        
        RobotController.__init__( self, dof )
        
        # Label
        self.name = 'Joint PID Controller'
        
        # Gains
        self.kp = np.ones( dof  ) * kp
        self.kd = np.ones( dof  ) * kd
        self.ki = np.ones( dof  ) * ki
        
        # TODO Update this is not a static controller !!!!
        # Integral Memory
        self.dt    = 0.001
        self.e_int = np.zeros( dof )
        
    
    #############################
    def c( self , y , r , t = 0 ):
        """ 
        Feedback static computation u = c(y,r,t)
        
        INPUTS
        y  : sensor signal vector     p x 1
        r  : reference signal vector  k x 1
        t  : time                     1 x 1
        
        OUTPUTSS
        u  : control inputs vector    m x 1
        
        """
        
        u = np.zeros(self.m) 
        
        # Ref
        q_d = r
        
        # Feedback from sensors
        x = y
        [ q , dq ] = self.x2q( x )
        
        # Error
        e  = q_d - q
        de =     - dq
        ie = self.e_int + e * self.dt # TODO use dynamic controller class
        
        # PIDs
        u = e * self.kp + de * self.kd + ie * self.ki
        
        # Memory
        self.e_int =  ie # TODO use dynamic controller class
        
        return u
    


###############################################################################
# End-Effector PID
###############################################################################
        
class old_EndEffectorPID( RobotController ) :
    """ 
    PID in effector coordinates, using the Jacobian of the system
    ---------------------------------------
    r  : reference signal vector  e   x 1
    y  : sensor signal vector     n   x 1
    u  : control inputs vector    dof x 1
    t  : time                     1   x 1
    ---------------------------------------
    u = c( y , r , t ) = (r - q) * kp + ( dq ) * kd + int(r - q) * ki

    """
    
    ############################
    def __init__(self, robot, kp = 1, ki = 0, kd = 0):
        """ """
        
        # Using from model
        self.fwd_kin = robot.forward_kinematic_effector
        self.J       = robot.J
        self.e       = robot.e # nb of effector dof
        
        RobotController.__init__( self, robot.dof )
        
        # Label
        self.name = 'End-Effector PID Controller'
        
        # Gains
        self.kp = np.ones( self.dof  ) * kp
        self.kd = np.ones( self.dof  ) * kd
        self.ki = np.ones( self.dof  ) * ki
        
        # TODO Update this is not a static controller !!!!
        # Integral Memory
        self.dt    = 0.001
        self.e_int = np.zeros( self.dof )
        
    
    #############################
    def c( self , y , r , t = 0 ):
        """ 
        Feedback static computation u = c(y,r,t)
        
        INPUTS
        y  : sensor signal vector     p x 1
        r  : reference signal vector  k x 1
        t  : time                     1 x 1
        
        OUTPUTS
        u  : control inputs vector    m x 1
        
        """
        
        u = np.zeros(self.m) 
        
        # Feedback from sensors
        x = y
        [ q , dq ] = self.x2q( x )
        
        # Jacobian computation
        J = self.J( q )
        
        # Ref
        r_desired   = r
        r_actual    = self.fwd_kin( q )
        dr          = np.dot( J , dq )
        
        # Error
        e  = r_desired - r_actual
        de =           - dr
        ie = self.e_int + e * self.dt # TODO use dynamic controller class
        
        # Effector space PID
        f = e * self.kp + de * self.kd + ie * self.ki
        
        # From effector force to joint torques
        u = np.dot( J.T , f )
        
        # Memory
        self.e_int =  ie # TODO use dynamic controller class
        
        return u
    
    
'''
#################################################################
##################          Main                         ########
#################################################################
'''


if __name__ == "__main__":
    pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May  4 10:01:30 2023

@author: alex
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic   import cartpole
from pyro.analysis  import costfunction
from pyro.control   import lqr
###############################################################################


# sys
sys  = cartpole.CartPole()

sys.xbar[1] = np.pi # Up-right position


# Cost function
cf  = costfunction.QuadraticCostFunction.from_sys( sys )

cf.Q[0,0] = 100.0
cf.Q[1,1] = 1.0
cf.R[0,0] = 100

ctl = lqr.linearize_and_synthesize_lqr_controller( sys , cf )


# Simulation
cl_sys = ctl + sys

cl_sys.x0[0] = -3.0
cl_sys.x0[1] = 2.5
    
cl_sys.compute_trajectory( tf = 10.0 )
cl_sys.plot_trajectory('xu')
cl_sys.animate_simulation( time_factor_video=1.0 )


# -*- coding: utf-8 -*-
"""
Created on Sun May 12 16:34:44 2019

@author: Alexandre
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.control  import robotcontrollers
from pyro.dynamic  import manipulator
###############################################################################


# Dynamic model (inputs are motor torques)
torque_controlled_robot    = manipulator.TwoLinkManipulator()

# Kinematic only model (inputs are motor velocities)
speed_controlled_robot  = manipulator.SpeedControlledManipulator.from_manipulator( torque_controlled_robot )


# Kinematic effector position controller
kinematic_controller = robotcontrollers.EndEffectorKinematicController( speed_controlled_robot )

# Controller params
kinematic_controller.gains = np.array([1.0,1.0]) 
kinematic_controller.rbar  = np.array([0.5,0.5]) # Goal position


# Closed-loop system
closed_loop_robot = kinematic_controller + speed_controlled_robot

# Initial position of the robot
closed_loop_robot.x0        = np.array([-0.5,0.2])

# Simulation and plots
closed_loop_robot.plot_trajectory('xu')
closed_loop_robot.animate_simulation()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May  1 19:51:49 2020

@author: alex
"""
import numpy as np

from gro640_robots import DrillingRobot
from abcd1234      import goal2r, r2q # Load your functions

# Define end-effector motion
r_0 = np.array([  0.5,   0.0,   1.0]) # start
r_f = np.array([ -0.25, -0.4,   0.6]) # end-point
t_f = 3.0                             # duration

# Compute the trajectory of the effector
r, dr, ddr = goal2r( r_0 , r_f , t_f )

n = r.shape[1] # Number of time steps

# Compute the trajectory of the joints
model      = DrillingRobot() # Robot Model
q, dq, ddq = r2q( r, dr, ddr , model  )

t   = np.linspace(0, t_f, n)   # t
q1  = q[0,:]
q2  = q[1,:]
q3  = q[2,:]
dq1 = dq[0,:]
dq2 = dq[1,:]
dq3 = dq[2,:]

###################################################
# Direct joint trajectory test
#q1 = np.linspace(0,1, n)     # q1(t)
#q2 = np.linspace(1.4,2, n)   # q2(t)
#q3 = np.linspace(-1.3,2, n)  # q3(t)
###################################################


###################################################
# Visualisation of the trajectory

sys = DrillingRobot()
sys.compute_trajectory( t_f , n ) # little hack to use the plotting tool without simulations

sys.traj.t       = t
sys.traj.x[:,0]  = q1
sys.traj.x[:,1]  = q2
sys.traj.x[:,2]  = q3
sys.traj.x[:,3]  = dq1
sys.traj.x[:,4]  = dq2
sys.traj.x[:,5]  = dq3

# Visualise trajectory with animation
sys.animate_simulation( is_3d = True )

# Visualise joint trajectory 
sys.plot_trajectory('x')

# Visualise x-y-z trajectory of the end-effector
sys.plot_end_effector_trajectory()


import numpy as np

from scipy import linalg

from pyro.dynamic  import ContinuousDynamicSystem
from pyro.analysis import simulation


###############################################################################
class StateSpaceSystem(ContinuousDynamicSystem):
    """Time-invariant state space representation of dynamic system

    f = A x + B u
    h = C x + D u

    Parameters
    ----------
    A, B, C, D : array_like
        The matrices which define the system

    """
    ############################################
    def __init__(self, A, B, C, D):
        
        self.A = A
        self.B = B
        self.C = C
        self.D = D

        self._check_dimensions()

        n = A.shape[1]
        m = B.shape[1]
        p = C.shape[0]
        
        ContinuousDynamicSystem.__init__( self, n, m, p)
        
        self.is_vectorized = True
        
    ############################################
    def _check_dimensions(self):
        
        if self.A.shape[0] != self.A.shape[1]:
            raise ValueError("A must be square")

        if self.B.shape[0] != self.A.shape[0]:
            raise ValueError("Number of rows in B does not match A")

        if self.C.shape[1] != self.A.shape[0]:
            raise ValueError("Number of columns in C does not match A")

        if self.D.shape[1] != self.B.shape[1]:
            raise ValueError("Number of columns in D does not match B")

        if self.C.shape[0] != self.D.shape[0]:
            raise ValueError("Number of rows in C does not match D")
    
    #############################################
    def f(self, x, u, t = 0 ):

        dx = np.dot(self.A, x) + np.dot(self.B, u)

        return dx
    
    #############################################
    def h(self, x, u, t = 0 ):
        
        y = np.dot(self.C, x) + np.dot(self.D, u)
        
        return y
    
    
    ############################################
    def compute_eigen_modes(self):
        
        D,V = linalg.eig( self.A )
        
        self.poles = D
        self.modes = V
        
        return (D,V)
    
    ############################################
    def compute_eigen_mode_traj(self, i = 0 ):
        """ 
        Simulation of time evolution of the system on mode i
        ------------------------------------------------
        i : mode index
        """
        
        #Time scaling for the mode
        norm = np.sqrt(self.poles[i].real**2 + self.poles[i].imag**2)
        
        if norm > 0.001:
            tf = 2. / norm * 2 * np.pi + 1
            tf = np.clip(tf , 1 , 30)
        else:
            tf = 5
            
        n  = 2001

        sim = simulation.Simulator(self, tf, n)
        
        sim.x0 = self.modes[:,i].real + self.xbar

        traj   = sim.compute() # save the result in the instance

        return traj
    
    
    ############################################
    def animate_eigen_mode(self, i = 0 , is_3d = False):
        """ 
        Simulation of time evolution of the system on mode i
        ------------------------------------------------
        i : mode index
        """
        
        # Compute eigen decomposition
        self.compute_eigen_modes()
        
        # Simulate one mode
        traj = self.compute_eigen_mode_traj( i )
        
        # Animate mode
        animator       = self.get_animator()
        
        template = 'Mode %i \n%0.1f+%0.1fj'
        label    = template % (i, self.poles[i].real, self.poles[i].imag)
        
        animator.top_right_label = label
        
        return animator.animate_simulation( traj, 3.0, is_3d)
        
        

    
    
    
    

################################################################
def _approx_jacobian(func, xbar, epsilons):
    """ Numerically approximate the jacobian of a function

    Parameters
    ----------
    func : callable
        Function for which to approximate the jacobian. Must accept an array of
        dimension ``n`` and return an array of dimension ``m``.
    xbar : array_like (dimension ``n``)
        Input around which the jacobian will be evaluated.
    epsilons : array_like (dimension ``n``)
        Step size to use for each input when approximating the jacobian

    Returns
    -------
    jac : array_like
        Jacobian matrix with dimensions m x n
    """

    n  = xbar.shape[0]
    ybar = func(xbar)
    m  = ybar.shape[0]

    J = np.zeros((m, n))
    
    for i in range(n):
        # Forward evaluation
        xf    = np.copy(xbar)
        xf[i] = xbar[i] + epsilons[i]
        yf    = func(xf)

        # Backward evaluation
        xb    = np.copy(xbar)
        xb[i] = xbar[i] - epsilons[i]
        yb    = func(xb)
        
        # Slope
        delta = yf - yb

        J[:, i] = delta / (2.0 * epsilons[i])

    return J


#################################################################
def linearize(sys, epsilon_x=0.001, epsilon_u=None):
    """Generate linear state-space model by linearizing any system.

    The system to be linearized is assumed to be time-invariant.

    Parameters
    ----------
    sys : `pyro.dynamic.ContinuousDynamicSystem`
        The system to linearize
    xbar : array_like
        State array arround which the system will be linearized
    epsilon : float
        Step size to use for numerical gradient approximation

    Returns
    -------
    instance of `StateSpaceSystem`

    """
    
    xbar = sys.xbar.astype(float)
    ubar = sys.ubar.astype(float)
    tbar = sys.tbar

    epsilon_x = np.asarray(epsilon_x)

    if epsilon_u is None:
        if epsilon_x.size > 1:
            raise ValueError("If epsilon_u is not provided, epsilon_x must be scalar")
        epsilon_u = epsilon_x

    epsilon_u = np.asarray(epsilon_u)

    if epsilon_u.size == 1:
        epsilon_u = np.ones(sys.m) * epsilon_u

    if epsilon_x.size == 1:
        epsilon_x = np.ones(sys.n) * epsilon_x
        

    def f_x(x):
        return sys.f(x, ubar, tbar)

    def f_u(u):
        return sys.f(xbar, u, tbar)

    def h_x(x):
        return sys.h(x, ubar, tbar)

    def h_u(u):
        return sys.h(xbar, u, tbar)

    A = _approx_jacobian(f_x, xbar, epsilon_x)
    B = _approx_jacobian(f_u, ubar, epsilon_u)
    C = _approx_jacobian(h_x, xbar, epsilon_x)
    D = _approx_jacobian(h_u, ubar, epsilon_u)
    
    ss = StateSpaceSystem(A, B, C, D)
    
    #############
    # Labels
    #############
    
    for i in range(sys.n):
        ss.state_label[i]  = 'Delta ' + sys.state_label[i]
    
    ss.state_units  = sys.state_units
    
    for i in range(sys.p):
        ss.output_label[i] = 'Delta ' + sys.output_label[i]
        
    ss.output_units = sys.output_units
    
    for i in range(sys.m):
        ss.input_label[i]  = 'Delta ' + sys.input_label[i]
        
    ss.input_units  = sys.input_units
    
    ss.name = 'Linearized ' + sys.name
    
    #############
    # Graphical
    #############
    
    # New fonction from delta_states to configuration space
    def new_xut2q( x, u, t):
        
        x = x + sys.xbar
        u = u + sys.ubar
        
        return sys.xut2q( x, u, t)
    
    ss.xut2q                     = new_xut2q
    
    # Using the non-linear sys graphical kinematic
    ss.linestyle                = sys.linestyle
    ss.forward_kinematic_domain = sys.forward_kinematic_domain
    ss.forward_kinematic_lines  = sys.forward_kinematic_lines

    return ss


class StateObserver(StateSpaceSystem):
    """Linear time-invariant continuous-time state observer

    f = d(x_est)/dt = A x_est + B u_sys + L(y_sys - y_est)

    Where x_est is the estimate of x, the state vector and y_est is the
    estimate of y, the output vector, based on x_est and u_sys.

    States of observer (n = n_sys):

        x = x_est

    Inputs of observer (m = m_sys + p_sys):

        u = [u_sys; y_sys]

    Outputs of observer (p = n = n_sys):

        y = x = x_est


    Parameters
    ----------
    A : array-like      n x n
        Systems dynamics (state transition) matrix of the observer plant model
    B : array-like      n x m
        Input matrix of the observer plant model
    C : array-like      p x n
        State-Output matrix of the observer plant model
    D : array-like      p x m
        Input-Output matrix of the observer plant model
    L:  array-like      n x p
        Observer gain matrix. n and p refer respectively to the number of states and the
        number of outputs of `sys`.

    """

    def __init__(self, A, B, C, D, L):
        self.A = np.array(A, ndmin=2, dtype=np.float64)
        self.B = np.array(B, ndmin=2, dtype=np.float64)
        self.C = np.array(C, ndmin=2, dtype=np.float64)
        self.D = np.array(D, ndmin=2, dtype=np.float64)
        self.L = np.array(L, ndmin=2, dtype=np.float64)


        self.m_plant = self.B.shape[1]
        self.p_plant = self.C.shape[0]

        # Observer states = Estimated states
        n = self.A.shape[1]

        # Observer inputs = concatenation of y and u of observed system
        m = self.m_plant + self.p_plant

        # Outputs of observer = estimated states
        p = n

        ContinuousDynamicSystem.__init__(self, n, m, p)

        self._check_dimensions()


    def _check_dimensions(self):
        super()._check_dimensions()

        # L must be n x p of sys
        if self.L.shape[0] != self.n or self.L.shape[1] != self.p_plant:
            raise ValueError("Dimensions of gain matrix L do not match system ss.")


    @classmethod
    def from_ss(cls, ss, L):
        """Create a state observer based on an existing state-space system"""
        obs = cls(ss.A, ss.B, ss.C, ss.D, L)

        obs.name = "Observer for " + ss.name

        obs.state_label = [f"Estimated {l}" for l in ss.state_label]
        obs.state_units = ss.state_units

        obs.input_label = ss.input_label + ss.output_label
        obs.input_units = ss.input_units + ss.output_units

        obs.output_label = obs.state_label
        obs.output_units = obs.state_units

        return obs


    @classmethod
    def kalman(cls, A, B, C, D, Q, R, G=None):
        """ Create a state observer by calculating the Kalman gain matrix.

        This method calculates the Kalman gain matrix L_Kalman for the system:

        dx/dt = Ax + Bu + Gw
        y = Cx + Du + v

        Where w and v are normally distributed random vectors with 0 mean and
        covariance matrices Q and V.

        Notes on matrix G:

            - In the case where the noise process w is additive onto the system inputs,
              dx/dt = Ax + B(u + w), we have B(u + w) = Bu + Bw and therefore `G` = `B`.
              This is the default case when `G` is left unspecified or `None`.

            - If the noise process is additive onto the system states,
              dx/dt = Ax + Bu + w, then `G=I` should be passed as an argument, where `I`
              is the identity matrix with the same shape as A (n x n).

        Parameters
        ----------

        A, B, C, D : array-like
            See description in `StateObserver` class.
        Q : array-like      q x q
            Covariance matrix of the process noise w (q x 1)
        R : array-like      p x p
            Covariance matrix of the output noise v (m x 1)
        G : array-like      n x q
            Input matrix for the noise process w. By default (`G=None`), it is assumed
            that the noise process w is additive on the input u, therefore G = B and
            q = m.

        Returns
        ----------

        Instance of `StateObserver` with L, the Kalman gain matrix. A special
        property `P` is set which corresponds to the state estimation
        covariance matrix.

        """
        Q = np.array(Q, ndmin=2, dtype=np.float64)
        R = np.array(R, ndmin=2, dtype=np.float64)

        if G is None:
            G = B
        else:
            G = np.array(G, ndmin=2, dtype=np.float64)

        L = np.zeros([A.shape[0], C.shape[0]]) # temporary
        obs = cls(A, B, C, D, L)

        # Check dimensions of Q, R and G matrices
        if not Q.shape[0] == Q.shape[1]:
            raise ValueError("Q must be square")
        if not R.shape[0] == R.shape[1]:
            raise ValueError("R must be square")
        if not G.shape[0] == A.shape[0]:
            raise ValueError("Shape[0] of G does not match shape of A")
        if not G.shape[1] == Q.shape[0]:
            raise ValueError("Shape[1] of G does not match shape of Q")
        if not R.shape[0] == C.shape[0]:
            raise ValueError("Shape of R must match number of outputs of C")

        P = linalg.solve_continuous_are(a=obs.A.T, b=obs.C.T, q=(G @ Q @ G.T), r=R)
        LT = np.linalg.solve(R.T, (obs.C @ P.T))
        if LT.ndim < 2:
            LT = LT[:, np.newaxis]
        L_kalm = LT.T
        assert L_kalm.shape == obs.L.shape

        obs.L = L_kalm
        obs.P = P # estimate covariance matrix
        return obs


    @classmethod
    def kalman_from_ss(cls, ss, Q, R, G=None):
        """Create a state observer by calculating the Kalman gain matrix.

        See documentation for `kalman(...)`. This method uses the A, B, C, D matrices
        from the system `ss`.

        Returns
        ----------

        Instance of `StateObserver` with L, the Kalman gain matrix. A special
        property `P` is set which corresponds to the state estimation
        covariance matrix.

        """

        kalm_obs = cls.kalman(ss.A, ss.B, ss.C, ss.D, Q, R, G)
        result = cls.from_ss(ss, kalm_obs.L)
        result.P = kalm_obs.P
        return result


    def h(self, x, u, t):
        # Output of observer system is the full vector of estimated states
        return x


    def f(self, x_est, u, t):

        # Check and adjust dimensions

        assert x_est.size == self.n
        x_est = x_est.reshape(self.n)

        assert u.size == self.m
        u = u.reshape(self.m)
        u_plant, y_plant = u[:self.m_plant], u[self.m_plant]

        assert y_plant.size == self.p_plant

        # State observer equations
        y_est = (self.C @ x_est) + (self.D @ u_plant)
        dx_est = (self.A @ x_est) + (self.B @ u_plant) + (self.L @ (y_plant - y_est))
        return dx_est.flatten()


    def __add__(self, sys):
        return ObservedSystem(sys, self)


class ObservedSystem(ContinuousDynamicSystem):
    """Combination of a dynamic system and a state observer

    States of observed system (n = 2 * n_sys):

        x = [x_sys; x_est]

    Inputs of observed system:

        u = u_sys

    Outputs of observed system:

        y = x_est

    """

    def __init__(self, sys, obs):
        self.sys = sys
        self.obs = obs

        n = sys.n * 2
        m = sys.m
        p = sys.n

        if not sys.n == obs.n:
            raise ValueError("Number of states of observer does not match system")

        if not sys.m == obs.m_plant:
            raise ValueError("Number of plant inputs of observer does not match system")

        if not sys.p == obs.p_plant:
            raise ValueError("Number of plant outputs of observer does not match system")

        super().__init__(n, m, p)

        self.x0 = np.concatenate([self.sys.x0, self.obs.x0], axis=0)

        self.state_label = sys.state_label + [f"Estimated {l}" for l in sys.state_label]
        self.state_units = sys.state_units + sys.state_units

        self.input_label = sys.input_label
        self.input_units = sys.input_units

        self.output_label = self.state_label[:sys.n]
        self.output_units = sys.state_units


    def f(self, x, u, t):
        u = np.array(u, ndmin=1).reshape(self.m)
        n_sys = self.sys.n
        x_sys, x_est = x[:n_sys], x[n_sys:]
        assert x_sys.shape == x_est.shape

        dx_sys = self.sys.f(x_sys, u, t)
        y_sys = self.sys.h(x_sys, u, t)

        u_obs = np.concatenate([u, y_sys], axis=0)
        dx_obs = self.obs.f(x_est, u_obs, t)

        return np.concatenate([dx_sys, dx_obs], axis=0)


    def h(self, x, u, t):
        n_sys = self.sys.n
        x_est = x[n_sys:]
        return x_est


    def t2u(self, t):
        return self.sys.t2u(t)



'''
#################################################################
##################          Main                         ########
#################################################################
'''


if __name__ == "__main__":     
    """ MAIN TEST """
    from pyro.dynamic import pendulum
    
    non_linear_sys = pendulum.SinglePendulum()
    non_linear_sys.xbar = np.array([0.,0.])
    
    EPS = 0.001
    
    linearized_sys = linearize( non_linear_sys , EPS )
    
    print('\nA:\n',linearized_sys.A)
    print('\nB:\n',linearized_sys.B)
    print('\nC:\n',linearized_sys.C)
    print('\nD:\n',linearized_sys.D)
    
    # Small oscillations
    non_linear_sys.x0 = np.array([0.1,0])
    linearized_sys.x0 = np.array([0.1,0])
    
    non_linear_sys.compute_trajectory()
    linearized_sys.compute_trajectory()
    
    non_linear_sys.plot_trajectory()
    linearized_sys.plot_trajectory()
    
    # Large oscillations
    non_linear_sys.x0 = np.array([1.8,0])
    linearized_sys.x0 = np.array([1.8,0])
    
    non_linear_sys.compute_trajectory()
    linearized_sys.compute_trajectory()
    
    non_linear_sys.plot_trajectory()
    linearized_sys.plot_trajectory()
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May  4 10:01:30 2023

@author: alex
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic   import pendulum
from pyro.analysis  import costfunction
from pyro.planning  import trajectoryoptimisation
from pyro.control   import lqr
from pyro.analysis  import simulation
###############################################################################

sys = pendulum.SinglePendulum()


# Cost function
cf  = costfunction.QuadraticCostFunction.from_sys( sys )

cf.Q[0,0] = 1.0
cf.Q[1,1] = 1.0
cf.R[0,0] = 0.1


#Max/Min torque
sys.u_ub[0] = +4
sys.u_lb[0] = -6

planner = trajectoryoptimisation.DirectCollocationTrajectoryOptimisation( sys )

planner.x_start = np.array([0,0])
planner.x_goal  = np.array([-3.14,0])

planner.init_dynamic_plot()
planner.compute_optimal_trajectory()

traj = planner.traj


cf.R[0,0] = 1.0


ctl = lqr.TrajectoryLQRController( sys , traj , cf )


# Simulation
cl_sys = ctl + sys

cl_sys.x0[0] = 0.0 + 0.2
cl_sys.x0[1] = 0.0
    
cl_sys.compute_trajectory( tf = 5.0 )
cl_sys.plot_trajectory('xu')
cl_sys.animate_simulation( time_factor_video=1.0 )




# -*- coding: utf-8 -*-
"""
Created on Mon Nov 12 20:28:17 2018

@author: Alexandre
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic  import cartpole
from pyro.planning import randomtree
###############################################################################


sys  = cartpole.UnderActuatedRotatingCartPole()

x_start = np.array([0,-3.14,0,0])
x_goal  = np.array([0,0,0,0])

planner = randomtree.RRT( sys , x_start )

t = 50
    
planner.u_options = [
        np.array([-t]),
        np.array([+t]),
        np.array([ 0])
        ]

planner.goal_radius          = 1.5
planner.max_nodes            = 10000
planner.max_solution_time    = 1.5
planner.dt                   = 0.05
planner.max_distance_compute = 1000
planner.dyna_plot            = False

planner.find_path_to_goal( x_goal )

planner.plot_tree()
planner.show_solution()

anim = planner.animate_solution( is_3d = True )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 30 14:04:51 2021

@author: alex
"""

import numpy as np

from pyro.dynamic.pendulum                import SinglePendulum
from pyro.planning.trajectoryoptimisation import DirectCollocationTrajectoryOptimisation


sys = SinglePendulum()

#Max/Min torque
sys.u_ub[0] = +4
sys.u_lb[0] = -6

planner = DirectCollocationTrajectoryOptimisation( sys )

planner.x_start = np.array([0,0])
planner.x_goal  = np.array([-3.14,0])

planner.compute_optimal_trajectory()
planner.show_solution()
planner.animate_solution()


# -*- coding: utf-8 -*-
"""
Created on Sun May 12 16:34:44 2019

@author: Alexandre
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.control  import robotcontrollers
from pyro.control  import nonlinear
from pyro.dynamic  import manipulator
###############################################################################

sys     = manipulator.TwoLinkManipulator()

# Estimated parameters
sys.l1      = 0.3
sys.l2      = 0.3
sys.lc1     = 0.3
sys.lc2     = 0.3
sys.I1      = 0.5
sys.I2      = 0.2
sys.m1      = 0.9
sys.m2      = 0.05
sys.u_lb[0] = -1.0
sys.u_lb[1] = -1.0
sys.u_ub[0] = 1.0
sys.u_ub[1] = 1.0
sys.l_domain = 0.6

# Target
q_desired = np.array([0.,0.])

# Joint PD
dof = 2
joint_pd      = robotcontrollers.JointPD( dof )
joint_pd.rbar = q_desired
joint_pd.kp   = np.array([ 3.0, 3.0 ])
joint_pd.kd   = np.array([ 1.0, 1.0 ])

# Effector PD
eff_pd      = robotcontrollers.EndEffectorPD( sys )
eff_pd.rbar = np.array([-0.4,+0.2])
eff_pd.kp   = np.array([ 3.0, 3.0 ])
eff_pd.kd   = np.array([ 1.0, 1.0 ])

## Computed torque controller
ct_ctl      = nonlinear.ComputedTorqueController( sys )
ct_ctl.w0   = 2.0
ct_ctl.zeta = 0.7
ct_ctl.rbar = np.array([0.0,0.0])

# Controller selection
ctl = joint_pd
ctl = ct_ctl
ctl = eff_pd

# Closed-loops
cl_sys    = ctl + sys

# Simulations
tf = 5
cl_sys.x0 = np.array([-3.14,0,0,0])
cl_sys.compute_trajectory( tf )
cl_sys.plot_trajectory('xu')
cl_sys.animate_simulation()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun 20 10:29:50 2023

@author: alex
"""

###############################################################################
import numpy as np
import matplotlib.pyplot as plt
###############################################################################
from pyro.analysis  import graphical
from pyro.dynamic   import mechanical
from pyro.kinematic import geometry
from pyro.kinematic import drawing
###############################################################################



##############################################################################
# 2D planar plane
##############################################################################
        
class Plane2D( mechanical.MechanicalSystemWithPositionInputs ):
    
    """ 
    Equations of Motion
    -------------------------
    
    """
    
    ############################
    def __init__(self):
        """ """
        
        # initialize standard params
        mechanical.MechanicalSystemWithPositionInputs.__init__( self, 3 , 1 , 1 )
        
        # Labels
        self.name = '2D plane model'
        self.state_label = ['x','y','theta','vx','vy','w']
        self.input_label = ['Trust', 'delta']
        self.output_label = self.state_label
        
        # Units
        self.state_units = ['[m]','[m]','[rad]','[m/sec]','[m/sec]','[rad/sec]']
        self.input_units = ['[N]', '[Rad]']
        self.output_units = self.state_units
        
        # State working range
        self.x_ub = np.array([+100,+200,+2,30,30,10])
        self.x_lb = np.array([-100,-0,-2,-30,-30,-10])
        
        self.u_ub = np.array([+10,+0.3])
        self.u_lb = np.array([ 0, -0.3])
        
        # Model param
        self.mass           = 2.0      # kg
        self.inertia        = 0.1       # kgm2
        self.gravity        = 9.8
        
        # Aero param
        self.rho            = 1.29      # air density
        
        self.S_w            = 0.2       # wing ref. area
        self.S_t            = 0.05      # tail ref. area
        self.l_w            = 0.0       # wing a.c. position with respect to c.g., negative is behind c.g.
        self.l_t            = 1.0       # tail a.c. position with respect to c.g., negative is behind c.g.
        
        # we assume same wing profile and geometry for wing and tail
        self.Cd0            = 0.02     # parasite drag
        self.AR             = 5.0      # aspect ratio
        self.e_factor       = 0.8      # oswald efficiency factor
        self.Cm0            = 0.0      # Aero moment coef. ac
        self.alpha_stall    = np.pi / 12.      
        
        
        # Graphic output parameters 
        self.length          = 2.0   
        self.l_cg            = self.length * 0.6 # distance from back of airplane to cg
        self.width           = self.length / 10.0
        self.dynamic_domain  = True
        self.dynamic_range   = self.length * 1.0
        self.static_range    = self.length * 30
        
        
    ###########################################################################
    def compute_velocity_vector(self, q , dq ):
        
        theta = q[2]
        vx    = dq[0]
        vy    = dq[1]
        
        V     = np.sqrt( vx**2 + vy**2 )  # absolute velocity
        gamma = np.arctan2( vy , vx )     # velocity vector angle
        
        alpha = theta - gamma             # angle of attack
        
        return ( V , gamma , alpha )
    
    
    ###########################################################################
    def Cl(self, alpha ):
        
        # Rough fit on
        # https://www.aerospaceweb.org/question/airfoils/q0150b.shtml
        
        Cl = np.sin( 2 * alpha ) # falt plate approx
        
        #If not stalled
        if (alpha < self.alpha_stall ) and (alpha > -self.alpha_stall ):
            
            Cl = Cl + 4 * alpha
        
        return Cl
    
    
    ###########################################################################
    def Cd(self, alpha ):
        
        Cl = self.Cl( alpha )
        
        # Body parasite drag
        Cd = self.Cd0 
        
        # Wing flat plate approx
        Cd = Cd + ( 1 - np.cos( 2 * alpha ))
        
        #If not stalled: add induced drag
        if (alpha < self.alpha_stall ) and (alpha > -self.alpha_stall ):
            
            Cd = Cd + Cl **2 / ( np.pi * self.e_factor * self.AR )
                
        
        return Cd
    
    
    ###########################################################################
    def Cm(self, alpha ):
        
        Cm = self.Cm0 
        
        return Cm
    
    
    #############################
    def plot_alpha2Cl(self, alpha_min = -3.15, alpha_max = 3.15 ):
        
        alphas = np.arange( alpha_min, alpha_max, 0.05 )
        
        n   = alphas.shape[0]
        Cls = np.zeros((n,1))
        Cds = np.zeros((n,1))
        Cms = np.zeros((n,1))
        
        for i in range(n):
            Cls[i] = self.Cl( alphas[i] )
            Cds[i] = self.Cd( alphas[i] )
            Cms[i] = self.Cm( alphas[i] )
        
        fig , ax = plt.subplots(3, figsize=graphical.default_figsize,
                                dpi= graphical.default_dpi, frameon=True)

        fig.canvas.manager.set_window_title('Aero curve')
        
        ax[0].plot( alphas , Cls , 'b')
        ax[0].set_ylabel('Cl', fontsize=graphical.default_fontsize)
        ax[0].set_xlabel('alpha', fontsize=graphical.default_fontsize )
        ax[0].tick_params( labelsize = graphical.default_fontsize )
        ax[0].grid(True)
        
        ax[1].plot( alphas , Cds , 'b')
        ax[1].set_ylabel('Cd', fontsize=graphical.default_fontsize)
        ax[1].set_xlabel('alpha', fontsize=graphical.default_fontsize )
        ax[1].tick_params( labelsize = graphical.default_fontsize )
        ax[1].grid(True)
        
        ax[2].plot( alphas , Cms , 'b')
        ax[2].set_ylabel('Cm', fontsize=graphical.default_fontsize)
        ax[2].set_xlabel('alpha', fontsize=graphical.default_fontsize )
        ax[2].tick_params( labelsize = graphical.default_fontsize )
        ax[2].grid(True)
        
        fig.tight_layout()
        fig.canvas.draw()
        
        plt.show()
    
    
    
    ###########################################################################
    def compute_aerodynamic_forces( self, V , alpha , delta ):
        
        rv2 = 0.5 * self.rho * V**2
        
        c_w = np.sqrt( self.S_w / self.AR ) 
        c_t = np.sqrt( self.S_t / self.AR ) 
        
        Cl_w = self.Cl( alpha )
        Cd_w = self.Cd( alpha )
        Cm_w = self.Cm( alpha )
        
        L_w = rv2 * self.S_w * Cl_w
        D_w = rv2 * self.S_w * Cd_w
        M_w = rv2 * self.S_w * c_w * Cm_w
        
        Cl_t = self.Cl( alpha + delta )
        Cd_t = self.Cd( alpha + delta )
        Cm_t = self.Cm( alpha + delta )
        
        L_t = rv2 * self.S_t * Cl_t
        D_t = rv2 * self.S_t * Cd_t
        M_t = rv2 * self.S_t * c_t * Cm_t
        
        return ( L_w , D_w , M_w , L_t , D_t , M_t )
        
        
        
    ###########################################################################
    def H(self, q ):
        """ 
        Inertia matrix 
        ----------------------------------
        dim( H ) = ( dof , dof )
        
        such that --> Kinetic Energy = 0.5 * dq^T * H(q) * dq
        
        """  
        
        H = np.zeros((3,3))
        
        H[0,0] = self.mass
        H[1,1] = self.mass
        H[2,2] = self.inertia
        
        return H
    
    
    ###########################################################################
    def C(self, q , dq ):
        """ 
        Corriolis and Centrifugal Matrix 
        ------------------------------------
        dim( C ) = ( dof , dof )
        
        such that --> d H / dt =  C + C^T
        
        
        """ 
        
        C = np.zeros((3,3))

        return C
    
    
    ###########################################################################
    def g(self, q ):
        """ 
        Gravitationnal forces vector : dof x 1
        """
        
        g = np.zeros(3)
        
        g[1] =  self.mass * self.gravity

        return g
    
    
    ###########################################################################
    def d(self, q , dq , u ):
        """ 
        State-dependent dissipative forces : dof x 1
        """
        
        V , gamma , alpha = self.compute_velocity_vector( q , dq )
        
        delta = u[1]
        
        L_w, D_w, M_w, L_t, D_t, M_t = self.compute_aerodynamic_forces( V , alpha, delta )
        
        ##########################################################
        # Total aero forces vector at c.g. in wind-aligned basis
        ##########################################################
        
        s = np.sin( alpha )
        c = np.cos( alpha )
        
        L = L_w + L_t
        D = D_w + D_t
        M = M_w + M_t - self.l_w * ( L_w * c + D_w * s ) - self.l_t * ( L_t * c + D_t * s )
        
        ##########################################################
        # Transformation of aero forces in global inertial basis
        ##########################################################
        
        d_wind = np.array([ -D , L , M ]) 
        
        s = np.sin( gamma )
        c = np.cos( gamma )
        
        R = np.array([ [ c   , -s ,  0 ] , 
                       [ s   ,  c ,  0 ] ,
                       [ 0   ,  0 ,  1 ]   ])
        
        d = - R @ d_wind # aero forces in global inertial basis
        
        return d
    
    ###########################################################################
    def B(self, q , u ):
        """ 
        Actuator Matrix  : dof x m
        """
        
        B = np.zeros((3,1))
        
        theta = q[2]
        
        # TODO PLACE HOLDER
        B[0,0] = np.cos( theta )
        B[1,0] = np.sin( theta )
        
        return B
    
    
    ###########################################################################
    def forward_kinematic_domain(self, q ):
        """ 
        """
        
        x = q[0] 
        y = q[1] 
        z = 0
        
        if self.dynamic_domain:
            
            l = self.dynamic_range
        
            domain  = [ ( -l + x , l + x ) ,
                        ( -l + y , l + y ) ,
                        ( -l + z , l + z ) ]
        else:
            
            l = self.static_range
            
            domain  = [ ( -l * 0.01 , l ) ,
                        ( -l * 0.01 , l ) ,
                        ( -l * 0.01 , l ) ]#
            
                
        return domain
    
    
    ###########################################################################
    def forward_kinematic_lines(self, q ):
        """ 
        Compute points p = [x;y;z] positions given config q 
        ----------------------------------------------------
        - points of interest for ploting
        
        Outpus:
        lines_pts = [] : a list of array (n_pts x 3) for each lines
        
        """
        
        lines_pts   = [] # list of array (n_pts x 3) for each lines
        lines_style = []
        lines_color = []
        
        
        ###########################
        # Dimensions
        ###########################
        
        w = self.width  # body width
        l = self.length # body lenght
        
        ###########################
        # Body
        ###########################
        
        pts      = np.zeros(( 5 , 3 ))
        
        x     = q[0]
        y     = q[1]
        theta = q[2]
        
        world_T_body = geometry.transformation_matrix_2D( theta , x , y )
        #body_T_wind  = transformation_matrix_2D_from_base_angle( -alpha , 0 , 0 )
        body_T_drawing = geometry.transformation_matrix_2D( 0 , -self.l_cg , -w/2 )
        
        body_pts_local = np.array([ [ 0   ,  0   ,  1 ] , 
                                    [ l   ,  0   ,  1 ] ,
                                    [ l-w ,  w   ,  1 ] ,
                                    [ 2*w ,  w   ,  1 ] ,
                                    [ w   ,  3*w ,  1 ] ,
                                    [ 0   ,  3*w ,  1 ] ,
                                    [ 0   ,  0   ,  1 ] ])

        
        body_pts_global = drawing.transform_points_2D( world_T_body @  body_T_drawing , body_pts_local )
        
        lines_pts.append( body_pts_global )
        lines_style.append( '-')
        lines_color.append( 'b')
        
        if self.dynamic_domain :
            
            cg = np.array([ [ x  ,  y   ,  1 ] ])
            
            lines_pts.append( cg )
            lines_style.append( 'o')
            lines_color.append( 'k')
        
        ###########################
        # Wings
        ###########################
        
        pts      = np.zeros(( 2 , 3 ))
        
        c_w = np.sqrt( self.S_w / self.AR )

        
        wings_pts_body = np.array([ [ -self.l_w + c_w   ,  0  ,  1 ] , 
                                    [ -self.l_w - c_w   ,  0  ,  1 ] ])
        
        
        wings_pts_world = drawing.transform_points_2D( world_T_body , wings_pts_body )
        
        lines_pts.append( wings_pts_world )
        lines_style.append( '-')
        lines_color.append( 'b')
        
        ###########################
        # bottom line
        ###########################
        
        pts = np.zeros((2,3))
        
        pts[0,0] = -10000
        pts[1,0] = 10000
        pts[0,1] = 0
        pts[1,1] = 0
        
        lines_pts.append( pts )
        lines_style.append('--')
        lines_color.append('k')
        
            
        return lines_pts , lines_style , lines_color
    
    
    ###########################################################################
    def forward_kinematic_lines_plus(self, x , u , t ):
        """ 
        plots the force vector
        
        """
        
        lines_pts = [] # list of array (n_pts x 3) for each lines
        lines_style = []
        lines_color = []
        
        #w = self.width
        
        q, dq = self.x2q(x)
        
        x     = q[0]
        y     = q[1]
        theta = q[2]
        
        V , gamma , alpha = self.compute_velocity_vector( q , dq )
        
        
        world_T_body = geometry.transformation_matrix_2D( theta , x , y )
        body_T_wind  = geometry.transformation_matrix_2D( -alpha , 0 , 0 )
        
        delta = u[1]
        
        ###########################
        # Trust vector
        ###########################
        
        # Max trust --> arrow is long as airplane
        f_scale = self.length / (self.u_ub[0] - self.u_lb[0])
        
        trust_vector_lenght = u[0] * f_scale
        
        #pts  = arrow_from_tip_angle( trust_vector_lenght , theta , bx , by )
        
        trust_arrow_body = drawing.arrow_from_length_angle( trust_vector_lenght, 0, -self.l_cg, 0 , origin = 'tip')
        
        trust_arrow_world = drawing.transform_points_2D( world_T_body , trust_arrow_body )
        
        lines_pts.append( trust_arrow_world )
        lines_style.append( '-')
        lines_color.append( 'r')
        
        ###########################
        # Control surface
        ###########################
        
        c_t = np.sqrt( self.S_t / self.AR )
        
        # NOT TO scale to better see the elevator
        tail_pts_tail = np.array([ [  1.0 * c_t   ,  0  ,  1 ] , 
                                   [ -1.0 * c_t   ,  0  ,  1 ] ]) 
        
        body_T_tail = geometry.transformation_matrix_2D( delta , - self.l_t , 0 )
        
        tail_pts_global = drawing.transform_points_2D( world_T_body @ body_T_tail , tail_pts_tail )
        
        lines_pts.append( tail_pts_global )
        lines_style.append( '-')
        lines_color.append( 'b')
        
        # ###########################
        # # Velocity vector
        # ###########################
        
        v_length = V * self.length / self.x_ub[3]
        if v_length > self.length: v_length = self.length
        
        v_pts = drawing.arrow_from_length_angle( v_length , 0 )
        
        v_world = drawing.transform_points_2D( world_T_body @ body_T_wind , v_pts )
        
        lines_pts.append( v_world  )
        lines_style.append('-')
        lines_color.append('k')
        
        # ###########################
        # # Aero forces
        # ###########################
        
        L_w, D_w, M_w, L_t, D_t, M_t = self.compute_aerodynamic_forces( V , alpha , delta )
        
        L_w_pts = drawing.arrow_from_components(               0, L_w * f_scale )
        D_w_pts = drawing.arrow_from_components(  -D_w * f_scale,             0 )
        L_t_pts = drawing.arrow_from_components(               0, L_t * f_scale )
        D_t_pts = drawing.arrow_from_components(  -D_t * f_scale,             0 )
        
        body_T_acw = geometry.transformation_matrix_2D( 0 , -self.l_w , 0  )
        body_T_act = geometry.transformation_matrix_2D( 0 , -self.l_t , 0  )
        
        L_w_pts_global = drawing.transform_points_2D( world_T_body @  body_T_acw @ body_T_wind , L_w_pts )
        D_w_pts_global = drawing.transform_points_2D( world_T_body @  body_T_acw @ body_T_wind , D_w_pts )
        
        lines_pts.append( L_w_pts_global )
        lines_style.append('-')
        lines_color.append('b')
        
        lines_pts.append( D_w_pts_global )
        lines_style.append('-')
        lines_color.append( 'r' )
        
        L_t_pts_global = drawing.transform_points_2D( world_T_body @  body_T_act @ body_T_wind , L_t_pts )
        D_t_pts_global = drawing.transform_points_2D( world_T_body @  body_T_act @ body_T_wind , D_t_pts )
        
        
        lines_pts.append( L_t_pts_global )
        lines_style.append('-')
        lines_color.append('b')
        
        lines_pts.append( D_t_pts_global )
        lines_style.append('-')
        lines_color.append('r')
        
                
        return lines_pts , lines_style , lines_color



    
'''
#################################################################
##################          Main                         ########
#################################################################
'''


if __name__ == "__main__":     
    """ MAIN TEST """
    
    
    if True:
    
        sys = Plane2D()
        
        #sys.plot_alpha2Cl()
        
        sys.x0   = np.array([0,0,0.2,15,0,0])
        
        
        
        def t2u(t):
            
            u = np.array([ 2 * t , -0.12 * t ])
            
            return u
            
        sys.t2u = t2u
        
        #sys.gravity = 0
        
        sys.compute_trajectory( 2 , 1001 , 'euler' )
        sys.plot_trajectory('x')
        
        # sys.dynamic_domain = False
        sys.animate_simulation( time_factor_video=0.5 )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Sep  8 10:38:30 2023

@author: alex
"""

import numpy as np

from pyro.dynamic import plane

sys = plane.Plane2D()

sys.plot_alpha2Cl()

sys.x0   = np.array([0,0,0.2,15,0,0])

sys.inertia = 1.0


def t2u(t):
    
    u = np.array([ 2 * t , -0.12 * t ])
    
    return u
    
sys.t2u = t2u

#sys.gravity = 0

sys.compute_trajectory( 10 , 20001 , 'euler' )
sys.plot_trajectory('x')

# sys.dynamic_domain = False
sys.animate_simulation( time_factor_video=0.5 )